def example(extract='m', estimate=False):
    import numpy as np
    import xarray as xr
    from addicty import Dict

    import larch as lx

    hh, pp, tour, skims = lx.example(200, ["hh", "pp", "tour", "skims"])

    exampville_mode_choice_file = lx.example(
        201, output_file="exampville_mode_choice.html", estimate=True
    )
    m = lx.load_model(exampville_mode_choice_file)

    Mode = Dict(
        DA=1,
        SR=2,
        Walk=3,
        Bike=4,
        Transit=5,
    ).freeze()

    tour_dataset = lx.Dataset.construct.from_idco(tour.set_index("TOURID"), alts=Mode)
    od_skims = lx.Dataset.construct.from_omx(skims)

    dt = lx.DataTree(
        tour=tour_dataset.dc.query_cases("TOURPURP == 1"),
        hh=hh.set_index("HHID"),
        person=pp.set_index("PERSONID"),
        od=od_skims,
        do=od_skims,
        relationships=(
            "tours.HHID @ hh.HHID",
            "tours.PERSONID @ person.PERSONID",
            "hh.HOMETAZ @ od.otaz",
            "tours.DTAZ @ od.dtaz",
            "hh.HOMETAZ @ do.dtaz",
            "tours.DTAZ @ do.otaz",
        ),
    )

    logsums = lx.DataArray.zeros(
        dt.caseids(),
        skims.TAZ_ID,
        name="logsums",
    )

    for dtaz in logsums.TAZ_ID:
        m.datatree = dt.replace_datasets(
            tour=dt.root_dataset.assign(DTAZ=xr.full_like(dt._getitem("DTAZ"), dtaz)),
        )
        logsums.loc[dict(TAZ_ID=dtaz)] = m.logsums()

    logsums.to_zarr("logsums.zarr")

    lx.DataArray.from_zarr("logsums.zarr")
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]

    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



from __future__ import annotations

import numpy as np

from ..util.simple_attribute import SimpleAttribute
from .basemodel import BaseModel
from .constraints import ParametricConstraintList


class ConstrainedModel(BaseModel):
    constraint_intensity = SimpleAttribute(float)
    constraint_sharpness = SimpleAttribute(float)
    constraints = ParametricConstraintList()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.constraint_intensity = 1.0
        self.constraint_sharpness = 1.0

        self._constraint_funcs = None
        """Functions for calculating constraint values."""

    def mangle(self, data=True, structure=True) -> None:
        super().mangle(data, structure)
        if structure:
            self._constraint_funcs = None

    def unmangle(self, force=False, structure_only=False):
        if not self._mangled and not force:
            return
        marker = f"_currently_unmangling_{__file__}"
        if getattr(self, marker, False):
            return
        try:
            setattr(self, marker, True)
            super().unmangle(force=force)
            if not structure_only:
                if self._constraint_funcs is None:
                    self._constraint_funcs = [
                        c.as_soft_penalty() for c in self.constraints
                    ]
        finally:
            delattr(self, marker)

    def constraint_violation(
        self,
        on_violation="raise",
        intensity_check=False,
    ):
        """
        Check if constraints are currently violated.

        Parameters
        ----------
        on_violation : {'raise', 'return'}
            If set to 'raise', an exception is raised if any model constraint,
            including any bound constraint, is violated.  Otherwise, this method
            returns a message describing the first constraint violation found, or
            an empty string if no violation are found.
        intensity_check : bool, default False
            If True, when the model's `constraint_intensity` attribute is set
            to zero, this function always returns OK (empty string).

        Returns
        -------
        str
            If no exception is raised, this method returns a message describing
            the first constraint violation found, or an empty string if no
            violation are found.
        """
        OK = ""
        if intensity_check and self.constraint_intensity == 0:
            return OK
        over_max = self.pvals > self.pmaximum
        if np.any(over_max):
            failure = np.where(over_max)[0][0]
            failure_message = (
                f"{self.pnames[failure]} over maximum "
                f"({self.pvals[failure]} > {self.pmaximum[failure]})"
            )
            if on_violation != "raise":
                return failure_message
            raise ValueError(failure_message)
        under_min = self.pvals < self.pminimum
        if np.any(under_min):
            failure = np.where(under_min)[0][0]
            failure_message = (
                f"{self.pnames[failure]} under minimum "
                f"({self.pvals[failure]} < {self.pminimum[failure]})"
            )
            if on_violation != "raise":
                return failure_message
            raise ValueError(failure_message)
        for c in self.constraints:
            if c.fun(self.pvals) < 0:
                failure_message = str(c)
                if on_violation != "raise":
                    return failure_message
                raise ValueError(failure_message)
        return OK



from __future__ import annotations

import logging
import pathlib
import warnings
from collections import namedtuple
from pathlib import Path
from typing import Literal

import numpy as np
import pandas as pd
import xarray as xr
from numba import boolean, guvectorize, njit
from numba import float32 as f32
from numba import float64 as f64
from numba import int8 as i8
from numba import int32 as i32
from numba import int64 as i64

from ..dataset import DataArray, Dataset, DataTree
from ..exceptions import MissingDataError
from ..model.possible_overspec import (
    PossibleOverspecification,
    compute_possible_overspecification,
)
from ..util import dictx
from ..util.simple_attribute import SimpleAttribute
from .basemodel import BaseModel as _BaseModel
from .numba_stream import ModelStreamer

if not list(Path(__file__).parent.glob("__pycache__/numbamodel.*.nbc")):
    warnings.warn(  ## Good news, everyone! ##  )
        "\n\n"
        "#### larch v6 is experimental, and not feature-complete ####\n"
        "the first time you import on a new system, this package will\n"
        "compile optimized binaries for your specific machine,  which\n"
        "may take a little while, please be patient ...\n",
        stacklevel=1,
    )

logger = logging.getLogger(__package__)


@njit(cache=True)
def minmax(x):
    maximum = x[0]
    minimum = x[0]
    for i in x[1:]:
        if i > maximum:
            maximum = i
        elif i < minimum:
            minimum = i
    return (minimum, maximum)


@njit(cache=True)
def outside_range(x, bottom, top):
    for i in x[:]:
        if i == -np.inf:
            continue
        if i > top:
            return True
        elif i < bottom:
            return True
    return False


@njit(error_model="numpy", fastmath=True, cache=True)
def utility_from_data_co(
    model_utility_co_alt,  # int input shape=[n_co_features]
    model_utility_co_param_scale,  # float input shape=[n_co_features]
    model_utility_co_param,  # int input shape=[n_co_features]
    model_utility_co_data,  # int input shape=[n_co_features]
    parameter_arr,  # float input shape=[n_params]
    holdfast_arr,  # float input shape=[n_params]
    array_av,  # int8 input shape=[n_alts]
    data_co,  # float input shape=[n_co_vars]
    utility_elem,  # float output shape=[n_alts]
    dutility_elem,  # float output shape=[n_alts, n_params]
):
    for i in range(model_utility_co_alt.shape[0]):
        altindex = model_utility_co_alt[i]
        param_value = parameter_arr[model_utility_co_param[i]]
        param_holdfast = holdfast_arr[model_utility_co_param[i]]
        if array_av[altindex]:
            if model_utility_co_data[i] == -1:
                utility_elem[altindex] += param_value * model_utility_co_param_scale[i]
                if not param_holdfast:
                    dutility_elem[altindex, model_utility_co_param[i]] += (
                        model_utility_co_param_scale[i]
                    )
            else:
                _temp = (
                    data_co[model_utility_co_data[i]] * model_utility_co_param_scale[i]
                )
                utility_elem[altindex] += _temp * param_value
                if not param_holdfast:
                    dutility_elem[altindex, model_utility_co_param[i]] += _temp


@njit(error_model="numpy", fastmath=True, cache=True)
def quantity_from_data_ca(
    model_q_ca_param_scale,  # float input shape=[n_q_ca_features]
    model_q_ca_param,  # int input shape=[n_q_ca_features]
    model_q_ca_data,  # int input shape=[n_q_ca_features]
    model_q_scale_param,  # int input scalar
    parameter_arr,  # float input shape=[n_params]
    holdfast_arr,  # float input shape=[n_params]
    array_av,  # int8 input shape=[n_alts]
    array_ca,  # float input shape=[n_alts, n_ca_vars]
    utility_elem,  # float output shape=[n_alts]
    dutility_elem,  # float output shape=[n_alts, n_params]
):
    n_alts = array_ca.shape[0]

    if model_q_scale_param[0] >= 0:
        scale_param_value = parameter_arr[model_q_scale_param[0]]
        scale_param_holdfast = holdfast_arr[model_q_scale_param[0]]
    else:
        scale_param_value = 1.0
        scale_param_holdfast = 1

    for j in range(n_alts):
        # if self._array_ce_reversemap is not None:
        #     if c >= self._array_ce_reversemap.shape[0] or j >= self._array_ce_reversemap.shape[1]:
        #         row = -1
        #     else:
        #         row = self._array_ce_reversemap[c, j]
        _row = -1

        if array_av[j]:  # and row != -1:
            if model_q_ca_param.shape[0]:
                for i in range(model_q_ca_param.shape[0]):
                    # if row >= 0:
                    #     _temp = self._array_ce[row, self.model_quantity_ca_data[i]]
                    # else:
                    _temp = (
                        array_ca[j, model_q_ca_data[i]]
                        * model_q_ca_param_scale[i]
                        * np.exp(parameter_arr[model_q_ca_param[i]])
                    )
                    utility_elem[j] += _temp
                    if not holdfast_arr[model_q_ca_param[i]]:
                        dutility_elem[j, model_q_ca_param[i]] += (
                            _temp * scale_param_value
                        )

                for i in range(model_q_ca_param.shape[0]):
                    if not holdfast_arr[model_q_ca_param[i]]:
                        dutility_elem[j, model_q_ca_param[i]] /= utility_elem[j]

                _tempsize = np.log(utility_elem[j])
                utility_elem[j] = _tempsize * scale_param_value
                if (model_q_scale_param[0] >= 0) and not scale_param_holdfast:
                    dutility_elem[j, model_q_scale_param[0]] += _tempsize

        else:
            utility_elem[j] = -np.inf


@njit(error_model="numpy", fastmath=True, cache=True)
def utility_from_data_ca(
    model_utility_ca_param_scale,  # int input shape=[n_u_ca_features]
    model_utility_ca_param,  # int input shape=[n_u_ca_features]
    model_utility_ca_data,  # int input shape=[n_u_ca_features]
    parameter_arr,  # float input shape=[n_params]
    holdfast_arr,  # float input shape=[n_params]
    array_av,  # int8 input shape=[n_alts]
    array_ca,  # float input shape=[n_alts, n_ca_vars]
    array_ce_data,  # float input shape=[n_casealts, n_ca_vars]
    array_ce_indices,  # int input shape=[n_casealts]
    array_ce_ptr,  # int input shape=[2]
    utility_elem,  # float output shape=[n_alts]
    dutility_elem,  # float output shape=[n_alts, n_params]
):
    n_alts = array_ca.shape[0]

    if array_ce_data.shape[0] > 0:
        j = 0
        for row in range(array_ce_ptr[0], array_ce_ptr[1]):
            while array_ce_indices[row] > j:
                # skipped alts are unavail
                utility_elem[j] = -np.inf
                j += 1
            # j = array_ce_indices[row] # happens naturally
            for i in range(model_utility_ca_param.shape[0]):
                _temp = array_ce_data[row, model_utility_ca_data[i]]
                _temp *= model_utility_ca_param_scale[i]
                utility_elem[j] += _temp * parameter_arr[model_utility_ca_param[i]]
                if not holdfast_arr[model_utility_ca_param[i]]:
                    dutility_elem[j, model_utility_ca_param[i]] += _temp
            j += 1
        while n_alts > j:
            # skipped alts are unavail
            utility_elem[j] = -np.inf
            j += 1
    else:
        for j in range(n_alts):
            row = -1

            if array_av[j]:
                for i in range(model_utility_ca_param.shape[0]):
                    if row >= 0:
                        _temp = 0.0  # _temp = self._array_ce[row, self.model_utility_ca_data[i]]
                    else:
                        _temp = array_ca[j, model_utility_ca_data[i]]
                    _temp *= model_utility_ca_param_scale[i]
                    utility_elem[j] += _temp * parameter_arr[model_utility_ca_param[i]]
                    if not holdfast_arr[model_utility_ca_param[i]]:
                        dutility_elem[j, model_utility_ca_param[i]] += _temp
            else:
                utility_elem[j] = -np.inf


def _type_signature(sig, precision=32):
    result = ()
    for s in sig:
        if s == "f":
            result += (f32[:],) if precision == 32 else (f64[:],)
        elif s == "F":
            result += (f32[:, :],) if precision == 32 else (f64[:, :],)
        elif s == "r":
            result += (f32,) if precision == 32 else (f64,)
        elif s == "i":
            result += (i32[:],)
        elif s == "j":
            result += (i64[:],)
        elif s == "I":
            result += (i32[:, :],)
        elif s == "b":
            result += (i8[:],)
        elif s == "S":
            result += (i8,)
        elif s == "B":
            result += (boolean,)
    return result


def _type_signatures(sig):
    return [
        _type_signature(sig, precision=32),
        _type_signature(sig, precision=64),
    ]


@njit(error_model="numpy", fastmath=True, cache=True)
def _numba_utility_to_loglike(
    n_alts,
    edgeslots,  # int input shape=[edges, 4]
    mu_slots,  # int input shape=[nests]
    start_slots,  # int input shape=[nests]
    len_slots,  # int input shape=[nests]
    holdfast_arr,  # int8 input shape=[n_params]
    parameter_arr,  # float input shape=[n_params]
    array_ch,  # float input shape=[nodes]
    array_av,  # int8 input shape=[nodes]
    array_wt,  # float input shape=[]
    return_flags,  #
    dutility,  #
    utility,  # float output shape=[nodes]
    logprob,  # float output shape=[nodes]
    probability,  # float output shape=[nodes]
    bhhh,  # float output shape=[n_params, n_params]
    d_loglike,  # float output shape=[n_params]
    loglike,  # float output shape=[]
):
    assert edgeslots.shape[1] == 4
    upslots = edgeslots[:, 0]  # int input shape=[edges]
    dnslots = edgeslots[:, 1]  # int input shape=[edges]
    visit1 = edgeslots[:, 2]  # int input shape=[edges]
    _allocslot = edgeslots[:, 3]  # int input shape=[edges]

    assert return_flags.size == 4
    only_utility = return_flags[0]  # [19] int8 input
    return_probability = return_flags[1]  # [20] bool input
    return_grad = return_flags[2]  # [21] bool input
    return_bhhh = return_flags[3]  # [22] bool input

    # util_nx = np.zeros_like(utility)
    # mu_extra = np.zeros_like(util_nx)
    loglike[0] = 0.0

    if True:  # outside_range(utility[:n_alts], -0.0, 0.0):
        for up in range(n_alts, utility.size):
            up_nest = up - n_alts
            n_children_for_parent = len_slots[up_nest]
            shifter = -np.inf
            shifter_position = -1
            if mu_slots[up_nest] < 0:
                mu_up = 1.0
            else:
                mu_up = parameter_arr[mu_slots[up_nest]]
            if mu_up:
                for n in range(n_children_for_parent):
                    edge = start_slots[up_nest] + n
                    dn = dnslots[edge]
                    if utility[dn] > -np.inf:
                        z = utility[dn] / mu_up
                        if z > shifter:
                            shifter = z
                            shifter_position = dn
                        # TODO alphas
                        # if alpha[edge] > 0:
                        #     z = (logalpha[edge] + utility[child]) / mu[parent]
                        #     if z > shifter:
                        #         shifter = z
                        #         shifter_position = child
                for n in range(n_children_for_parent):
                    edge = start_slots[up_nest] + n
                    dn = dnslots[edge]
                    if utility[dn] > -np.inf:
                        if shifter_position == dn:
                            utility[up] += 1
                        else:
                            utility[up] += np.exp((utility[dn] / mu_up) - shifter)
                        # if alpha[edge] > 0:
                        #     if shifter_position == child:
                        #         utility[parent] += 1
                        #     else:
                        #         z = ((logalpha[edge] + utility[child]) / mu[parent]) - shifter
                        #         utility[parent] += exp(z)
                utility[up] = (np.log(utility[up]) + shifter) * mu_up
            else:  # mu_up is zero
                for n in range(n_children_for_parent):
                    edge = start_slots[up_nest] + n
                    dn = dnslots[edge]
                    if utility[dn] > utility[up]:
                        utility[up] = utility[dn]
    else:
        for s in range(upslots.size):
            dn = dnslots[s]
            up = upslots[s]
            up_nest = up - n_alts
            dn_nest = dn - n_alts
            if mu_slots[up_nest] < 0:
                mu_up = 1.0
            else:
                mu_up = parameter_arr[mu_slots[up_nest]]
            if visit1[s] > 0 and dn >= n_alts:
                log_dn = np.log(utility[dn])
                # mu_extra[dn] += log_dn + util_nx[dn]/utility[dn]
                if mu_slots[dn_nest] < 0:
                    mu_dn = 1.0
                else:
                    mu_dn = parameter_arr[mu_slots[dn_nest]]
                utility[dn] = log_dn * mu_dn
            util_dn = utility[dn]
            exp_util_dn_mu_up = np.exp(util_dn / mu_up)
            utility[up] += exp_util_dn_mu_up
            # util_nx[up] -= util_dn * exp_util_dn_mu_up / mu_up

        # mu_extra[mu_extra.size-1] += np.log(utility[utility.size-1])
        #                              + util_nx[-1]/utility[utility.size-1]
        utility[utility.size - 1] = np.log(utility[utility.size - 1])

    if only_utility == 2:
        return

    for s in range(upslots.size):
        dn = dnslots[s]
        up = upslots[s]
        if mu_slots[up - n_alts] < 0:
            mu_up = 1.0
        else:
            mu_up = parameter_arr[mu_slots[up - n_alts]]
        if np.isinf(utility[up]) and utility[up] < 0:
            logprob[dn] = -np.inf
        else:
            logprob[dn] = (utility[dn] - utility[up]) / mu_up
        if array_ch[dn]:
            loglike[0] += logprob[dn] * array_ch[dn] * array_wt[0]

    if return_probability or return_grad or return_bhhh:
        # logprob becomes conditional_probability
        conditional_probability = logprob
        for i in range(logprob.size):
            if array_av[i]:
                conditional_probability[i] = np.exp(logprob[i])

        # probability
        probability[-1] = 1.0
        for s in range(upslots.size - 1, -1, -1):
            dn = dnslots[s]
            if array_av[dn]:
                up = upslots[s]
                probability[dn] = probability[up] * conditional_probability[dn]
            else:
                probability[dn] = 0.0

        if return_grad or return_bhhh:
            d_loglike[:] = 0.0

            # d utility
            for s in range(upslots.size):
                dn = dnslots[s]
                up = upslots[s]
                if array_av[dn]:
                    cond_prob = conditional_probability[dn]
                    if dn >= n_alts:
                        dn_mu_slot = mu_slots[dn - n_alts]
                        if dn_mu_slot >= 0:
                            dutility[dn, dn_mu_slot] += utility[dn]
                            dutility[dn, dn_mu_slot] /= parameter_arr[dn_mu_slot]
                    up_mu_slot = mu_slots[up - n_alts]
                    if up_mu_slot >= 0:
                        # FIXME: alpha slots to appear here if cross-nesting is activated
                        dutility[up, up_mu_slot] -= cond_prob * (utility[dn])
                    dutility[up, :] += np.where(
                        cond_prob, cond_prob * dutility[dn, :], 0
                    )

            # d probability
            # scratch = np.zeros_like(parameter_arr)
            # d_probability = np.zeros_like(dutility)
            # for s in range(upslots.size-1, -1, -1):
            #     dn = dnslots[s]
            #     if array_ch[dn]:
            #         up = upslots[s]
            #         scratch[:] = dutility[dn] - dutility[up]
            #         up_mu_slot = mu_slots[up-n_alts]
            #         if up_mu_slot < 0:
            #             mu_up = 1.0
            #         else:
            #             mu_up = parameter_arr[up_mu_slot]
            #         if mu_up:
            #             if up_mu_slot >= 0:
            #                 scratch[up_mu_slot] += (utility[up] - utility[dn]) / mu_up
            #                 # FIXME: alpha slots to appear here if cross-nesting is activated
            #             multiplier = probability[up] / mu_up
            #         else:
            #             multiplier = 0
            #
            #         scratch[:] *= multiplier
            #         scratch[:] += d_probability[up, :]
            #         d_probability[dn, :] += scratch[:] *
            #           conditional_probability[dn] # FIXME: for CNL, use edge not dn

            # d probability alternate path slightly lower memory usage and some faster
            d_probability = np.zeros_like(dutility)
            for s in range(upslots.size - 1, -1, -1):
                dn = dnslots[s]
                if array_ch[dn]:
                    up = upslots[s]
                    up_mu_slot = mu_slots[up - n_alts]
                    if up_mu_slot < 0:
                        mu_up = 1.0
                    else:
                        mu_up = parameter_arr[up_mu_slot]
                    for p in range(parameter_arr.size):
                        if mu_up:
                            scratch_ = dutility[dn, p] - dutility[up, p]
                            if p == up_mu_slot:
                                scratch_ += (utility[up] - utility[dn]) / mu_up
                                # FIXME: alpha slots to appear here if cross-nesting is activated
                            scratch_ *= probability[up] / mu_up
                        else:
                            scratch_ = 0
                        scratch_ += d_probability[up, p]
                        if conditional_probability[dn]:
                            d_probability[dn, p] += (
                                scratch_ * conditional_probability[dn]
                            )  # FIXME: for CNL, use edge not dn

            if return_bhhh:
                bhhh[:] = 0.0

            # d loglike
            for a in range(n_alts):
                this_ch = array_ch[a]
                if this_ch == 0:
                    continue
                total_probability_a = probability[a]
                # if total_probability_a > 0:
                #     tempvalue = d_probability[a, :] / total_probability_a
                #     if return_bhhh:
                #         bhhh += np.outer(tempvalue,tempvalue) * this_ch * array_wt[0]
                #     d_loglike += tempvalue * array_wt[0]
                #
                if total_probability_a > 0:
                    if total_probability_a < 1e-250:
                        total_probability_a = 1e-250
                    tempvalue = d_probability[a, :] * (this_ch / total_probability_a)
                    dLL_temp = tempvalue / this_ch
                    d_loglike += tempvalue * array_wt[0]
                    if return_bhhh:
                        bhhh += np.outer(dLL_temp, dLL_temp) * this_ch * array_wt[0]


_master_shape_signature = (
    "(qca),(qca),(qca),(), "
    "(uca),(uca),(uca), "
    "(uco),(uco),(uco),(uco), "
    "(edges,four), "
    "(nests),(nests),(nests), "
    "(params),(params), "
    "(nodes),(nodes),(),(vco),(alts,vca), "
    "(ces,vce),(ces),(two),  "
    "(four)->"
    "(nodes),(nodes),(nodes),(params,params),(params),()"
)


def _numba_master(
    model_q_ca_param_scale,  # [0] float input shape=[n_q_ca_features]
    model_q_ca_param,  # [1] int input shape=[n_q_ca_features]
    model_q_ca_data,  # [2] int input shape=[n_q_ca_features]
    model_q_scale_param,  # [3] int input scalar
    model_utility_ca_param_scale,  # [4] float input shape=[n_u_ca_features]
    model_utility_ca_param,  # [5] int input shape=[n_u_ca_features]
    model_utility_ca_data,  # [6] int input shape=[n_u_ca_features]
    model_utility_co_alt,  # [ 7] int input shape=[n_co_features]
    model_utility_co_param_scale,  # [ 8] float input shape=[n_co_features]
    model_utility_co_param,  # [ 9] int input shape=[n_co_features]
    model_utility_co_data,  # [10] int input shape=[n_co_features]
    edgeslots,  # [11] int input shape=[edges, 4]
    mu_slots,  # [12] int input shape=[nests]
    start_slots,  # [13] int input shape=[nests]
    len_slots,  # [14] int input shape=[nests]
    holdfast_arr,  # [15] int8 input shape=[n_params]
    parameter_arr,  # [16] float input shape=[n_params]
    array_ch,  # [17] float input shape=[nodes]
    array_av,  # [18] int8 input shape=[nodes]
    array_wt,  # [19] float input shape=[]
    array_co,  # [20] float input shape=[n_co_vars]
    array_ca,  # [21] float input shape=[n_alts, n_ca_vars]
    array_ce_data,  # [22] float input shape=[n_casealts, n_ca_vars]
    array_ce_indices,  # [23] int input shape=[n_casealts]
    array_ce_ptr,  # [24] int input shape=[2]
    return_flags,
    # only_utility,        # [19] int8 input
    # return_probability,  # [20] bool input
    # return_grad,         # [21] bool input
    # return_bhhh,         # [22] bool input
    utility,  # [23] float output shape=[nodes]
    logprob,  # [24] float output shape=[nodes]
    probability,  # [25] float output shape=[nodes]
    bhhh,  # [26] float output shape=[n_params, n_params]
    d_loglike,  # [27] float output shape=[n_params]
    loglike,  # [28] float output shape=[]
):
    n_alts = array_ca.shape[0]

    # assert edgeslots.shape[1] == 4
    # upslots   = edgeslots[:,0]  # int input shape=[edges]
    # dnslots   = edgeslots[:,1]  # int input shape=[edges]
    # visit1    = edgeslots[:,2]  # int input shape=[edges]
    # allocslot = edgeslots[:,3]  # int input shape=[edges]

    assert return_flags.size == 4
    only_utility = return_flags[0]  # int8 input
    # return_probability = return_flags[1]    # bool input
    # return_grad = return_flags[2]           # bool input
    # return_bhhh = return_flags[3]           # bool input

    utility[:] = 0.0
    dutility = np.zeros((utility.size, parameter_arr.size), dtype=utility.dtype)

    quantity_from_data_ca(
        model_q_ca_param_scale,  # float input shape=[n_q_ca_features]
        model_q_ca_param,  # int input shape=[n_q_ca_features]
        model_q_ca_data,  # int input shape=[n_q_ca_features]
        model_q_scale_param,  # int input scalar
        parameter_arr,  # float input shape=[n_params]
        holdfast_arr,  # float input shape=[n_params]
        array_av,  # int8 input shape=[n_nodes]
        array_ca,  # float input shape=[n_alts, n_ca_vars]
        utility[:n_alts],  # float output shape=[n_alts]
        dutility[:n_alts],
    )

    if only_utility == 3:
        if model_q_scale_param[0] >= 0:
            scale_param_value = parameter_arr[model_q_scale_param[0]]
        else:
            scale_param_value = 1.0
        utility[:n_alts] = np.exp(utility[:n_alts] / scale_param_value)
        return

    utility_from_data_ca(
        model_utility_ca_param_scale,  # int input shape=[n_u_ca_features]
        model_utility_ca_param,  # int input shape=[n_u_ca_features]
        model_utility_ca_data,  # int input shape=[n_u_ca_features]
        parameter_arr,  # float input shape=[n_params]
        holdfast_arr,  # float input shape=[n_params]
        array_av,  # int8 input shape=[n_nodes]
        array_ca,  # float input shape=[n_alts, n_ca_vars]
        array_ce_data,  # float input shape=[n_casealts, n_ca_vars]
        array_ce_indices,  # int input shape=[n_casealts]
        array_ce_ptr,  # int input shape=[2]
        utility[:n_alts],  # float output shape=[n_alts]
        dutility[:n_alts],
    )

    utility_from_data_co(
        model_utility_co_alt,  # int input shape=[n_co_features]
        model_utility_co_param_scale,  # float input shape=[n_co_features]
        model_utility_co_param,  # int input shape=[n_co_features]
        model_utility_co_data,  # int input shape=[n_co_features]
        parameter_arr,  # float input shape=[n_params]
        holdfast_arr,  # float input shape=[n_params]
        array_av,  # int8 input shape=[n_nodes]
        array_co,  # float input shape=[n_co_vars]
        utility[:n_alts],  # float output shape=[n_alts]
        dutility[:n_alts],
    )

    if only_utility == 1:
        return

    _numba_utility_to_loglike(
        n_alts,
        edgeslots,  # int input shape=[edges, 4]
        mu_slots,  # int input shape=[nests]
        start_slots,  # int input shape=[nests]
        len_slots,  # int input shape=[nests]
        holdfast_arr,  # int8 input shape=[n_params]
        parameter_arr,  # float input shape=[n_params]
        array_ch,  # float input shape=[nodes]
        array_av,  # int8 input shape=[nodes]
        array_wt,  # float input shape=[]
        return_flags,
        dutility,
        utility,  # float output shape=[nodes]
        logprob,  # float output shape=[nodes]
        probability,  # float output shape=[nodes]
        bhhh,  # float output shape=[n_params, n_params]
        d_loglike,  # float output shape=[n_params]
        loglike,  # float output shape=[]
    )


_numba_master_vectorized = guvectorize(
    _type_signatures("fiii fii ifii I iii bf fbffF Fij b fffFff"),
    _master_shape_signature,
    nopython=True,
    fastmath=True,
    target="parallel",
    cache=True,
)(
    _numba_master,
)


@njit(cache=True)
def softplus(i, sharpness=10):
    cut = 10 / sharpness
    if i > cut:
        return i
    # elif i < -cut:
    #     return 0.0
    else:
        return np.log1p(np.exp(i * sharpness)) / sharpness


@njit(cache=True)
def d_softplus(i, sharpness=10):
    cut = 1000 / sharpness
    if i > cut:
        return 1.0
    # elif i < -cut:
    #     return 0.0
    else:
        return 1 / (1 + np.exp(-i * sharpness))


@guvectorize(
    [
        _type_signature("ffffffff", precision=32),
        _type_signature("ffffffff", precision=64),
    ],
    ("(params),(params),(params),(),()->(),(params),(params)"),
    cache=True,
    nopython=True,
)
def bounds_penalty(
    param_array,  # [] float input shape=[n_params]
    lower_bounds,  # [] float input shape=[n_params]
    upper_bounds,  # [] float input shape=[n_params]
    constraint_intensity,  # [] float input shape=[]
    constraint_sharpness,  # [] float input shape=[]
    penalty,  # [] float output shape=[]
    d_penalty,  # [] float output shape=[n_params]
    d_penalty_binding,  # [] float output shape=[n_params]
):
    # penalty = 0.0
    # d_penalty = np.zeros_like(param_array)
    # d_penalty_binding = np.zeros_like(param_array)
    penalty[0] = 0.0
    d_penalty[:] = 0.0
    d_penalty_binding[:] = 0.0
    for i in range(param_array.size):
        diff_threshold = np.minimum(upper_bounds[i] - lower_bounds[i], 0.0001)
        low_diff = lower_bounds[i] - param_array[i]
        low_penalty = -softplus(low_diff, constraint_sharpness[0])
        high_diff = param_array[i] - upper_bounds[i]
        high_penalty = -softplus(high_diff, constraint_sharpness[0])
        penalty[0] += (low_penalty + high_penalty) * constraint_intensity[0]
        if low_penalty:
            d_penalty[i] += (
                d_softplus(lower_bounds[i] - param_array[i], constraint_sharpness[0])
                * constraint_intensity[0]
            )
        if high_penalty:
            d_penalty[i] -= (
                d_softplus(param_array[i] - upper_bounds[i], constraint_sharpness[0])
                * constraint_intensity[0]
            )
        if np.absolute(high_diff) < diff_threshold:
            d_penalty_binding[i] -= 0.5 * constraint_intensity[0]
        elif np.absolute(low_diff) < diff_threshold:
            d_penalty_binding[i] += 0.5 * constraint_intensity[0]
    # return penalty, d_penalty, d_penalty_binding


def _numba_penalty(
    param_array,  # [] float input shape=[n_params]
    lower_bounds,  # [] float input shape=[n_params]
    upper_bounds,  # [] float input shape=[n_params]
    constraint_intensity,  # [] float input shape=[]
    constraint_sharpness,  # [] float input shape=[]
    bhhh,  # [] float output shape=[n_params, n_params]
    d_loglike,  # [] float output shape=[n_params]
    loglike,  # [] float output shape=[]
):
    penalty = 0.0
    d_penalty = np.zeros_like(d_loglike)
    for i in range(param_array.size):
        low_penalty = -softplus(
            lower_bounds[i] - param_array[i], constraint_sharpness[0]
        )
        high_penalty = -softplus(
            param_array[i] - upper_bounds[i], constraint_sharpness[0]
        )
        penalty += (low_penalty + high_penalty) * constraint_intensity[0]
        if low_penalty:
            d_penalty[i] += (
                d_softplus(lower_bounds[i] - param_array[i], constraint_sharpness[0])
                * constraint_intensity[0]
            )
        if high_penalty:
            d_penalty[i] -= (
                d_softplus(param_array[i] - upper_bounds[i], constraint_sharpness[0])
                * constraint_intensity[0]
            )
    loglike[0] += penalty
    d_loglike[:] += d_penalty
    bhhh[:] += np.outer(d_penalty, d_penalty)


_numba_penalty_vectorized = guvectorize(
    _type_signatures("fffffFff"),
    ("(params),(params),(params),(),()->(params,params),(params),()"),
    nopython=True,
    fastmath=True,
    target="parallel",
    cache=True,
)(
    _numba_penalty,
)


def model_co_slots(data_provider: Dataset, model: _BaseModel, dtype=np.float64):
    len_co = sum(len(_) for _ in model.utility_co.values())
    model_utility_co_alt = np.zeros([len_co], dtype=np.int32)
    model_utility_co_param_scale = np.ones([len_co], dtype=dtype)
    model_utility_co_param = np.zeros([len_co], dtype=np.int32)
    model_utility_co_data = np.zeros([len_co], dtype=np.int32)

    j = 0

    param_loc = {}
    for _n, _pname in enumerate(model.pnames):
        param_loc[_pname] = _n
    data_loc = {}
    if isinstance(data_provider, Dataset):
        if "var_co" in data_provider.indexes:
            for _n, _dname in enumerate(data_provider.indexes["var_co"]):
                data_loc[_dname] = _n
        alternative_codes = data_provider.indexes[data_provider.dc.ALTID]
    else:
        raise TypeError(
            f"data_provider must be DataFrames or Dataset not {type(data_provider)}"
        )

    for alt, func in model.utility_co.items():
        altindex = alternative_codes.get_loc(alt)
        for i in func:
            model_utility_co_alt[j] = altindex
            model_utility_co_param[j] = param_loc[str(i.param)]
            model_utility_co_param_scale[j] = i.scale
            if i.data == "1":
                model_utility_co_data[j] = -1
            else:
                model_utility_co_data[j] = data_loc[
                    str(i.data)
                ]  # self._data_co.columns.get_loc(str(i.data))
            j += 1

    return (
        model_utility_co_alt,
        model_utility_co_param_scale,
        model_utility_co_param,
        model_utility_co_data,
    )


def model_u_ca_slots(data_provider: Dataset, model: _BaseModel, dtype=np.float64):
    if isinstance(data_provider, Dataset):
        looker = lambda tag: data_provider.indexes["var_ca"].get_loc(str(tag))
    else:
        raise TypeError(
            f"data_provider must be DataFrames or Dataset not {type(data_provider)}"
        )
    len_model_utility_ca = len(model.utility_ca)
    model_utility_ca_param_scale = np.ones([len_model_utility_ca], dtype=dtype)
    model_utility_ca_param = np.zeros([len_model_utility_ca], dtype=np.int32)
    model_utility_ca_data = np.zeros([len_model_utility_ca], dtype=np.int32)
    for n, i in enumerate(model.utility_ca):
        model_utility_ca_param[n] = model.get_param_loc(i.param)
        model_utility_ca_data[n] = looker(i.data)
        model_utility_ca_param_scale[n] = i.scale
    return (
        model_utility_ca_param_scale,
        model_utility_ca_param,
        model_utility_ca_data,
    )


def model_q_ca_slots(data_provider: Dataset, model: _BaseModel, dtype=np.float64):
    if isinstance(data_provider, Dataset):
        looker = lambda tag: data_provider.indexes["var_ca"].get_loc(str(tag))
    else:
        raise TypeError(
            f"data_provider must be DataFrames or Dataset not {type(data_provider)}"
        )
    len_model_q_ca = len(model.quantity_ca)
    model_q_ca_param_scale = np.ones([len_model_q_ca], dtype=dtype)
    model_q_ca_param = np.zeros([len_model_q_ca], dtype=np.int32)
    model_q_ca_data = np.zeros([len_model_q_ca], dtype=np.int32)
    if model.quantity_scale:
        model_q_scale_param = np.zeros([1], dtype=np.int32) + model.get_param_loc(
            model.quantity_scale
        )
    else:
        model_q_scale_param = np.zeros([1], dtype=np.int32) - 1
    for n, i in enumerate(model.quantity_ca):
        model_q_ca_param[n] = model.get_param_loc(i.param)
        model_q_ca_data[n] = looker(i.data)
        model_q_ca_param_scale[n] = i.scale
    return (
        model_q_ca_param_scale,
        model_q_ca_param,
        model_q_ca_data,
        model_q_scale_param,
    )


class _case_slice:
    def __get__(self, obj, objtype=None):
        self.parent = obj
        return self

    def __getitem__(self, idx):
        return type(self.parent)(
            **{k: getattr(self.parent, k)[idx] for k in self.parent._fields}
        )


WorkArrays = namedtuple(
    "WorkArrays",
    ["utility", "logprob", "probability", "bhhh", "d_loglike", "loglike"],
)
WorkArrays.cs = _case_slice()


FixedArrays = namedtuple(
    "FixedArrays",
    [
        "qca_scale",
        "qca_param_slot",
        "qca_data_slot",
        "qscale_param_slot",
        "uca_scale",
        "uca_param_slot",
        "uca_data_slot",
        "uco_alt_slot",
        "uco_scale",
        "uco_param_slot",
        "uco_data_slot",
        "edge_slots",
        "mu_slot",
        "start_edges",
        "len_edges",
    ],
)


class NumbaModel(_BaseModel):
    _null_slice = (None, None, None)
    streaming = ModelStreamer()
    constraint_intensity = SimpleAttribute(float)
    constraint_sharpness = SimpleAttribute(float)

    def __init__(self, *args, float_dtype=np.float64, datatree=None, **kwargs):
        for a in args:
            if datatree is None and isinstance(a, (DataTree, Dataset)):
                datatree = a
        super().__init__(datatree=datatree, **kwargs)
        self._dataset = None
        self._fixed_arrays = None
        self._data_arrays = None
        self.work_arrays = None
        self.float_dtype = float_dtype
        self.constraint_intensity = 0.0
        self.constraint_sharpness = 0.0
        self._constraint_funcs = None
        self.datatree = datatree
        self._should_preload_data = True
        cache_dir = kwargs.get("cache_dir", None)
        if cache_dir is not None:
            cache_dir = pathlib.Path(cache_dir)
            cache_dir.mkdir(exist_ok=True)
            self.datatree.cache_dir = cache_dir

        # when the alternative codes are set, and include zero, we need to
        # make change the root node id to be not zero.
        if isinstance(self.datatree, (DataTree, Dataset)):
            if 0 in self.datatree.dc.altids():
                # check that -1 is not in the altids
                if -1 in self.datatree.dc.altids():
                    raise ValueError("Cannot have both -1 and 0 as alternative codes")
                self.initialize_graph(root_id=-1)

    def save(self, filename, format="yaml", overwrite=False):
        from .saving import save_model

        return save_model(self, filename, format=format, overwrite=overwrite)

    def dumps(self):
        return repr(self.save(None, format="raw"))

    def should_preload_data(self, should=True):
        should = bool(should)
        if should and not self._should_preload_data:
            # not currently, but want to, mangle to prevent inconsistency
            self.mangle()
        self._should_preload_data = should

    @classmethod
    def from_dict(cls, content):
        self = cls()

        def loadthis(attr, wrapper=None, injector=None):
            i = content.get(attr, None)
            if i is not None:
                try:
                    if wrapper is not None:
                        i = wrapper(i)
                except AttributeError:
                    pass
                else:
                    if injector is None:
                        setattr(self, attr, i)
                    else:
                        injector(i)

        loadthis("float_dtype", lambda i: getattr(np, i))
        loadthis("compute_engine")
        loadthis("index_name")
        loadthis("parameters", xr.Dataset.from_dict, self.update_parameters)
        loadthis("availability_any")
        loadthis("availability_ca_var")
        loadthis("availability_co_vars")
        loadthis("choice_any")
        loadthis("choice_ca_var")
        loadthis("choice_co_code")
        loadthis("choice_co_vars")
        loadthis("common_draws")
        loadthis("constraint_intensity")
        loadthis("constraint_sharpness")
        loadthis("constraints")
        from .tree import NestingTree

        loadthis("graph", NestingTree.from_dict)
        loadthis("groupid")
        loadthis("logsum_parameter")
        loadthis("mixtures", self.mixtures.from_list)
        loadthis("n_draws")
        loadthis("prerolled_draws")
        loadthis("quantity_ca")
        loadthis("quantity_scale")
        loadthis("title")
        loadthis("utility_ca")
        loadthis("utility_co")
        loadthis("weight_co_var")
        loadthis("weight_normalization")
        return self

    work_arrays = SimpleAttribute()

    def mangle(self, data=True, structure=True) -> None:
        super().mangle(data, structure)
        if data:
            self._dataset = None
            self._data_arrays = None
            self.work_arrays = None
            self._array_ch_cascade = None
            self._array_av_cascade = None
        if structure:
            self._constraint_funcs = None
            self._fixed_arrays = None
            # print("self.streaming", self.streaming)
            self.streaming = None

    def is_mnl(self):
        """
        Check if this model is a MNL model.

        Returns
        -------
        bool
        """
        if self._graph is None:
            return True
        if len(self._graph) - len(self._graph.elementals) == 1:
            return True
        return False

    dataflows = SimpleAttribute(dict)

    def reflow_data_arrays(self) -> None:
        """Reload the internal data_arrays so they are consistent with the datatree."""
        if self.graph is None:
            self._data_arrays = None
            return

        if not self.use_streaming:
            logger.debug("Model.reflow_data_arrays with full datatree")
            datatree = self.datatree
        else:
            logger.debug(
                "Model.reflow_data_arrays with partial datatree, one case only"
            )
            datatree = self.datatree.replace_datasets(
                {
                    self.datatree.root_node_name: self.datatree.root_dataset.isel(
                        {self.datatree.CASEID: slice(0, 1)}
                    )
                }
            )
        if datatree is not None:
            from .data_arrays import prepare_data

            logger.debug(f"Model.datatree.cache_dir = {datatree.cache_dir}")
            self.dataset, self.dataflows = prepare_data(
                datasource=datatree,
                request=self,
                float_dtype=self.float_dtype,
                cache_dir=datatree.cache_dir,
                flows=self.dataflows,
                make_unused_flows=self.use_streaming,
            )
            if self.use_streaming:
                # when streaming the dataset created above is a vestigial
                # one-case dataset, really we just want the flows, so we
                # get rid of the dataset now
                self._dataset = None
                self._data_arrays = None
            else:
                if self.autoscale_weights:
                    self.dataset.dc.autoscale_weights()
                self._rebuild_data_arrays()
                if self.work_arrays is not None:
                    self._rebuild_work_arrays()

    def _rebuild_data_arrays(self):
        if self._dataset is None:
            self.reflow_data_arrays()  # create the dataset from the datatree
        else:
            self._data_arrays = self._dataset.dc.to_arrays(
                self.graph,
                float_dtype=self.float_dtype,
            )

    def _rebuild_work_arrays(
        self, n_cases=None, n_nodes=None, n_params=None, on_missing_data="silent"
    ):
        log = logging.getLogger("Larch")
        if n_cases is None:
            try:
                n_cases = self.n_cases
            except MissingDataError:
                if on_missing_data != "silent":
                    log.error("MissingDataError, cannot rebuild work arrays")
                self.work_arrays = None
                if on_missing_data == "raise":
                    raise
                return
        if n_nodes is None:
            n_nodes = len(self.graph)
        if n_params is None:
            n_params = self.n_params
        _need_to_rebuild_work_arrays = True
        if self.work_arrays is not None:
            if (
                (self.work_arrays.utility.shape[0] == n_cases)
                and (self.work_arrays.utility.shape[1] == n_nodes)
                and (self.work_arrays.d_loglike.shape[1] == n_params)
                and (self.work_arrays.utility.dtype == self.float_dtype)
            ):
                _need_to_rebuild_work_arrays = False
        if _need_to_rebuild_work_arrays:
            log.debug("rebuilding work arrays")
            self.work_arrays = WorkArrays(
                utility=np.zeros([n_cases, n_nodes], dtype=self.float_dtype),
                logprob=np.zeros([n_cases, n_nodes], dtype=self.float_dtype),
                probability=np.zeros([n_cases, n_nodes], dtype=self.float_dtype),
                bhhh=np.zeros([n_cases, n_params, n_params], dtype=self.float_dtype),
                d_loglike=np.zeros([n_cases, n_params], dtype=self.float_dtype),
                loglike=np.zeros([n_cases], dtype=self.float_dtype),
            )

    def _rebuild_fixed_arrays(self):
        data_provider = self.data_as_loaded
        if data_provider is not None:
            (
                model_utility_ca_param_scale,
                model_utility_ca_param,
                model_utility_ca_data,
            ) = model_u_ca_slots(data_provider, self, dtype=self.float_dtype)
            (
                model_utility_co_alt,
                model_utility_co_param_scale,
                model_utility_co_param,
                model_utility_co_data,
            ) = model_co_slots(data_provider, self, dtype=self.float_dtype)
            (
                model_q_ca_param_scale,
                model_q_ca_param,
                model_q_ca_data,
                model_q_scale_param,
            ) = model_q_ca_slots(data_provider, self, dtype=self.float_dtype)
            node_slot_arrays = self.graph.node_slot_arrays(self)
            n_alts = self.graph.n_elementals()
            self._fixed_arrays = FixedArrays(
                model_q_ca_param_scale,
                model_q_ca_param,
                model_q_ca_data,
                # (np.asarray([model_q_scale_param])
                # if self.use_streaming
                # else model_q_scale_param),
                model_q_scale_param,
                model_utility_ca_param_scale,
                model_utility_ca_param,
                model_utility_ca_data,
                model_utility_co_alt,
                model_utility_co_param_scale,
                model_utility_co_param,
                model_utility_co_data,
                np.stack(self.graph.edge_slot_arrays()).T,
                node_slot_arrays[0][n_alts:],
                node_slot_arrays[1][n_alts:],
                node_slot_arrays[2][n_alts:],
            )
        else:
            self._fixed_arrays = None

    def doctor(self, **kwargs):
        """
        Run diagnostics, checking for common problems and inconsistencies.

        See :func:`larch.model.troubleshooting.doctor` for more information.
        """
        result = super().doctor(**kwargs)
        self._rebuild_data_arrays()
        return result

    def unmangle(self, force=False, structure_only=False):
        if not self._mangled and not force:
            return
        marker = f"_currently_unmangling_{__file__}"
        if getattr(self, marker, False):
            return
        try:
            setattr(self, marker, True)
            super().unmangle(force=force)
            if not structure_only:
                if self._dataset is None or force:
                    self.reflow_data_arrays()
                if self._fixed_arrays is None or force:
                    self._rebuild_fixed_arrays()
                    self._rebuild_work_arrays()
                if self._constraint_funcs is None:
                    self._constraint_funcs = [
                        c.as_soft_penalty() for c in self.constraints
                    ]
        finally:
            delattr(self, marker)

    def _scan_logsums_ensure_names(self):
        nameset = set()
        try:
            g = self._graph
        except ValueError:
            pass
        else:
            if g is not None:
                for nodecode in g.topological_sorted_no_elementals:
                    if nodecode != g._root_id:
                        param_name = str(g.nodes[nodecode]["parameter"])
                        nameset.add(str(param_name))
        if self.quantity_ca is not None and len(self.quantity_ca) > 0:
            if self.quantity_scale is not None:
                nameset.add(str(self.quantity_scale))
        if self.logsum_parameter is not None:
            nameset.add(str(self.logsum_parameter))
        self._ensure_names(
            nameset, value=1, nullvalue=1, initvalue=1, minimum=0.01, maximum=1
        )

    def __prepare_for_compute(
        self,
        x=None,
        allow_missing_ch=False,
        allow_missing_av=False,
        caseslice=None,
    ):
        if caseslice is None:
            caseslice = slice(caseslice)
        if self.datatree is None and self.dataset is None:
            raise MissingDataError("dataset and datatree are both not set")
        self.unmangle()
        if x is not None:
            self.pvals = x
        if self.dataset is not None:
            if "ch" not in self.dataset and not allow_missing_ch:
                raise MissingDataError("model.dataset does not include `ch`")
        if self.work_arrays is None:
            self._rebuild_work_arrays(on_missing_data="raise")
        return (
            *self._fixed_arrays,
            self.pholdfast,
            self.pvals.astype(self.float_dtype),  # float input shape=[n_params]
            *self._data_arrays.cs[caseslice],  # TODO fix when not using named tuple
        )

    def constraint_violation(
        self,
        on_violation="raise",
        intensity_check=False,
    ):
        """
        Check if constraints are currently violated.

        Parameters
        ----------
        on_violation : {'raise', 'return'}
            If set to 'raise', an exception is raised if any model constraint,
            including any bound constraint, is violated.  Otherwise, this method
            returns a message describing the first constraint violation found, or
            an empty string if no violation are found.
        intensity_check : bool, default False
            If True, when the model's `constraint_intensity` attribute is set
            to zero, this function always returns OK (empty string).

        Returns
        -------
        str
            If no exception is raised, this method returns a message describing
            the first constraint violation found, or an empty string if no
            violation are found.
        """
        OK = ""
        if intensity_check and self.constraint_intensity == 0:
            return OK
        over_max = self.pvals > self.pmaximum
        if np.any(over_max):
            failure = np.where(over_max)[0][0]
            failure_message = (
                f"{self.pnames[failure]} over maximum "
                f"({self.pvals[failure]} > {self.pmaximum[failure]})"
            )
            if on_violation != "raise":
                return failure_message
            raise ValueError(failure_message)
        under_min = self.pvals < self.pminimum
        if np.any(under_min):
            failure = np.where(under_min)[0][0]
            failure_message = (
                f"{self.pnames[failure]} under minimum "
                f"({self.pvals[failure]} < {self.pminimum[failure]})"
            )
            if on_violation != "raise":
                return failure_message
            raise ValueError(failure_message)
        for c in self.constraints:
            if c.fun(self.pvals) < 0:
                failure_message = str(c)
                if on_violation != "raise":
                    return failure_message
                raise ValueError(failure_message)
        return OK

    def fit_bhhh(self, *args, **kwargs):
        from .numba_optimization import fit_bhhh

        return fit_bhhh(self, *args, **kwargs)

    def constraint_penalty(self, x=None):
        if x is not None:
            self.pvals = x
        penalty, dpenalty, dpenalty_binding = bounds_penalty(
            self.pvals.astype(self.float_dtype),
            self.pminimum.astype(self.float_dtype),
            self.pmaximum.astype(self.float_dtype),
            self.float_dtype(self.constraint_intensity),
            self.float_dtype(self.constraint_sharpness),
        )
        for cf, dcf, dcf_bind in self._constraint_funcs:
            penalty += cf(
                self.pvals,
                self.constraint_intensity,
                self.constraint_sharpness,
            )
            dpenalty += dcf(
                self.pvals,
                self.constraint_intensity,
                self.constraint_sharpness,
            )
            dpenalty_binding += dcf_bind(
                self.pvals,
                self.constraint_intensity,
            )
        return penalty, dpenalty, dpenalty_binding

    def constraint_converge_tolerance(self, x=None):
        args = self.__prepare_for_compute(
            x,
            allow_missing_ch=False,
        )
        args_flags = args + (
            np.asarray(
                [
                    0,  # only_utility
                    False,  # return_probability
                    True,  # return_gradient
                    True,  # return_bhhh
                ],
                dtype=np.int8,
            ),
        )
        with np.errstate(
            divide="ignore",
            over="ignore",
        ):
            _numba_master_vectorized(
                *args_flags,
                out=tuple(self.work_arrays),
            )
            if self.constraint_intensity:
                penalty, dpenalty, dpenalty_binding = self.constraint_penalty()
                self.work_arrays.loglike[:] += penalty
                self.work_arrays.d_loglike[:] += np.expand_dims(dpenalty_binding, 0)
                self.work_arrays.bhhh[:] = np.einsum(
                    "ij,ik->ijk", self.work_arrays.d_loglike, self.work_arrays.d_loglike
                )
        bhhh = self.work_arrays.bhhh.sum(0)
        dloglike = self.work_arrays.d_loglike.sum(0)
        freedoms = self.pholdfast == 0
        from .numba_optimization import propose_direction

        direction = propose_direction(bhhh, dloglike, freedoms)
        tolerance = np.dot(direction, dloglike) - self.n_cases
        return tolerance

    def _loglike_runner(
        self,
        x=None,
        only_utility=0,
        return_gradient=False,
        return_probability=False,
        return_bhhh=False,
        start_case=None,
        stop_case=None,
        step_case=None,
    ):
        caseslice = slice(start_case, stop_case, step_case)
        args = self.__prepare_for_compute(
            x,
            allow_missing_ch=return_probability or (only_utility > 0),
            caseslice=caseslice,
        )
        args_flags = args + (
            np.asarray(
                [
                    only_utility,
                    return_probability,
                    return_gradient,
                    return_bhhh,
                ],
                dtype=np.int8,
            ),
        )
        try:
            with np.errstate(
                divide="ignore",
                over="ignore",
            ):
                try:
                    result_arrays = WorkArrays(
                        *_numba_master_vectorized(
                            *args_flags,
                            out=tuple(self.work_arrays.cs[caseslice]),
                        )
                    )
                except ValueError:
                    result_arrays = WorkArrays(
                        *_numba_master_vectorized(
                            *args_flags,
                            # out=tuple(self.work_arrays.cs[caseslice]),
                        )
                    )

                if self.constraint_intensity:
                    penalty, dpenalty, dpenalty_binding = self.constraint_penalty()
                    self.work_arrays.loglike[caseslice] += np.nan_to_num(penalty)
                    self.work_arrays.d_loglike[caseslice] += np.nan_to_num(
                        np.expand_dims(dpenalty, 0)
                    )
                    self.work_arrays.bhhh[caseslice] = np.nan_to_num(
                        np.einsum(
                            "ij,ik->ijk",
                            self.work_arrays.d_loglike[caseslice],
                            self.work_arrays.d_loglike[caseslice],
                        )
                    )
                else:
                    penalty = 0.0

        except Exception:
            shp = lambda y: getattr(y, "shape", "scalar")
            dtp = lambda y: getattr(y, "dtype", f"{type(y)} ")
            import inspect

            arg_names = list(inspect.signature(_numba_master).parameters)
            arg_name_width = max(len(j) for j in arg_names)

            in_sig, out_sig = _master_shape_signature.split("->")
            in_sig_shapes = in_sig.split("(")[1:]
            out_sig_shapes = out_sig.split("(")[1:]
            print(in_sig_shapes)
            print(out_sig_shapes)
            print("# Input Arrays")
            for n, (a, s) in enumerate(zip(args_flags, in_sig_shapes)):
                s = s.rstrip(" ),")
                print(
                    f" {arg_names[n]:{arg_name_width}} [{n:2}] {s.strip():9}: {dtp(a)}{shp(a)}"
                )
            print("# Output Arrays")
            final_n = n
            for n, (a, s) in enumerate(
                zip(self.work_arrays, out_sig_shapes), start=final_n + 1
            ):
                s = s.rstrip(" ),")
                print(
                    f" {arg_names[n]:{arg_name_width}} [{n:2}] {s.strip():9}: {dtp(a)}{shp(a)}"
                )
            raise
        return result_arrays, penalty

    @property
    def weight_normalization(self):
        # try:
        #     return self.dataframes.weight_normalization
        # except AttributeError:
        return 1.0

    @weight_normalization.setter
    def weight_normalization(self, x):
        pass  # TODO

    def loglike(
        self,
        x=None,
        *,
        start_case: int | None = None,
        stop_case: int | None = None,
        step_case: int | None = None,
        check_if_best: bool = True,
        error_if_bad: bool = True,
        **kwargs,
    ):
        """
        Compute the log likelihood of the model.

        Parameters
        ----------
        x : array-like or dict, optional
            New values to set for the parameters before evaluating
            the log likelihood.  If given as array-like, the array must
            be a vector with length equal to the length of the
            parameter frame, and the given vector will replace
            the current values.  If given as a dictionary,
            the dictionary is used to update the parameters.
        start_case : int, default 0
            The first case to include in the log likelihood computation.
            To include all cases, start from 0 (the default).
        stop_case : int, default -1
            One past the last case to include in the log likelihood
            computation.  This is processed as usual for Python slicing
            and iterating, and negative values count backward from the
            end.  To include all cases, end at -1 (the default).
        step_case : int, default 1
            The step size of the case iterator to use in likelihood
            calculation.  This is processed as usual for Python slicing
            and iterating.  To include all cases, step by 1 (the default).
        check_if_best : bool, default True
            If True, check if the current log likelihood is the best
            found so far, and if so, update the cached best log likelihood
            and cached best parameters.
        error_if_bad : bool, default True
            If True, raise an exception if the log likelihood is NaN or Inf.

        Returns
        -------
        float
        """
        if self._use_streaming:
            result = (
                self.streaming.loglike(
                    x, start_case=start_case, stop_case=stop_case, step_case=step_case
                )
                * self.weight_normalization
            )
        else:
            result_arrays, penalty = self._loglike_runner(
                x, start_case=start_case, stop_case=stop_case, step_case=step_case
            )
            result = result_arrays.loglike.sum() * self.weight_normalization
            if error_if_bad:
                for f, tag in [(np.isnan, "NaN"), (np.isinf, "Inf")]:
                    if f(result):
                        if f(self.weight_normalization):
                            raise ValueError(f"weight_normalization is {tag}")
                        msg = f"log likelihood is {tag}"
                        bad_case_indexes = np.where(f(result_arrays.loglike))[0]
                        if len(bad_case_indexes) > 0:
                            msg += (
                                f" in {len(bad_case_indexes)} cases, including CASEIDs:"
                            )
                            caseids = self.dataset.dc.caseids()
                            msg += f" {caseids[bad_case_indexes[0]]}"
                            for i in bad_case_indexes[1:5]:
                                msg += f", {caseids[i]}"
                            if len(bad_case_indexes) > 5:
                                msg += ", ..."
                        raise ValueError(msg)
        if start_case is None and stop_case is None and step_case is None:
            self._check_if_best(result)
        return result

    def d_loglike(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        return_series=False,
        **kwargs,
    ):
        if self._use_streaming:
            result = (
                self.streaming.d_loglike(
                    x, start_case=start_case, stop_case=stop_case, step_case=step_case
                )
                * self.weight_normalization
            )
        else:
            result_arrays, penalty = self._loglike_runner(
                x,
                start_case=start_case,
                stop_case=stop_case,
                step_case=step_case,
                return_gradient=True,
            )
            result = result_arrays.d_loglike.sum(0) * self.weight_normalization
        if return_series:
            result = pd.Series(result, index=self.pnames)
        return result

    def loglike_casewise(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        **kwargs,
    ):
        result_arrays, penalty = self._loglike_runner(
            x,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
        )
        return result_arrays.loglike * self.weight_normalization

    def loglike_problems(self) -> pd.DataFrame:
        """
        Identify cases with log likelihood problems.

        Returns
        -------
        pandas.DataFrame
            A DataFrame identifying the caseids of cases with log likelihood
            problems, and the nature of the problem (e.g. NaN, +Inf, -Inf,
            exactly zero).
        """
        llcase = self.loglike_casewise()
        caseids = self.datatree.caseids()
        problems = pd.Series(np.nan, index=caseids, name="problem")
        problems.iloc[np.where(np.isnan(llcase))] = "nan"
        problems.iloc[np.where(np.isposinf(llcase))] = "+inf"
        problems.iloc[np.where(np.isneginf(llcase))] = "-inf"
        problems.iloc[np.where(llcase == 0)] = "zero"
        return problems.reset_index().rename_axis("caseindex").dropna()

    def loglike_null(self, use_cache=True):
        """
        Compute the log likelihood at null values.

        Set all parameter values to the value indicated in the
        "nullvalue" column of the parameter frame, and compute
        the log likelihood with the currently loaded data.  Note
        that the null value for each parameter may not be zero
        (for example, the default null value for logsum parameters
        in a nested logit model is 1).

        Parameters
        ----------
        use_cache : bool, default True
            Use the cached value if available.  Set to -1 to
            raise an exception if there is no cached value.

        Returns
        -------
        float
        """
        if self._cached_loglike_null is not None and use_cache:
            return self._cached_loglike_null
        elif use_cache == -1:
            raise ValueError("no cached value")
        else:
            current_parameters = self.pvals.copy()
            self.pvals = "null"
            self._cached_loglike_null = self.loglike()
            self.pvals = current_parameters
            return self._cached_loglike_null

    def d_loglike_casewise(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        **kwargs,
    ):
        result_arrays, penalty = self._loglike_runner(
            x,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
            return_gradient=True,
        )
        return result_arrays.d_loglike * self.weight_normalization

    def bhhh(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        return_dataframe=False,
        **kwargs,
    ):
        result_arrays, penalty = self._loglike_runner(
            x,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
            return_bhhh=True,
        )
        result = result_arrays.bhhh.sum(0) * self.weight_normalization
        if return_dataframe:
            result = pd.DataFrame(result, columns=self.pnames, index=self.pnames)
        return result

    def robust_covariance(self) -> xr.DataArray:
        """
        Compute the robust covariance matrix of the parameter estimates.

        Returns
        -------
        xarray.DataArray
        """
        if "robust_covariance_matrix" in self.parameters:
            return self.parameters["robust_covariance_matrix"]
        cov_dataarray = self.parameters["covariance_matrix"]
        covariance_matrix = cov_dataarray.to_numpy()
        take = np.full_like(covariance_matrix, True, dtype=bool)
        dense_s = len(self.pvals)
        for i in range(dense_s):
            if self.pholdfast[i]:
                # or (self.pf.loc[ii, 'value'] >= self.pf.loc[ii, 'maximum'])
                # or (self.pf.loc[ii, 'value'] <= self.pf.loc[ii, 'minimum']):
                take[i, :] = False
                take[:, i] = False
                dense_s -= 1
        if dense_s > 0:
            covariance_taken = covariance_matrix[take].reshape(dense_s, dense_s)
            # try:
            #     invhess = np.linalg.inv(hess_taken)
            # except np.linalg.linalg.LinAlgError:
            #     invhess = np.full_like(hess_taken, np.nan, dtype=np.float64)
            # robust...
            try:
                bhhh_taken = self.bhhh()[take].reshape(dense_s, dense_s)
            except NotImplementedError:
                robust_covariance_matrix = np.full_like(
                    covariance_matrix, 0, dtype=np.float64
                )
            else:
                # import scipy.linalg.blas
                # temp_b_times_h = scipy.linalg.blas.dsymm(float(1), invhess, bhhh_taken)
                # robusto = scipy.linalg.blas.dsymm(float(1), invhess, temp_b_times_h, side=1)
                robusto = np.dot(np.dot(covariance_taken, bhhh_taken), covariance_taken)
                robust_covariance_matrix = np.full_like(
                    covariance_matrix, 0, dtype=np.float64
                )
                robust_covariance_matrix[take] = robusto.reshape(-1)
        else:
            robust_covariance_matrix = np.full_like(
                covariance_matrix, 0, dtype=np.float64
            )

        self.add_parameter_array("robust_covariance_matrix", robust_covariance_matrix)
        self.add_parameter_array(
            "robust_std_err", _safe_sqrt(np.diagonal(robust_covariance_matrix))
        )
        return self.parameters["robust_covariance_matrix"]

    def _wrap_as_dataframe(
        self,
        arr,
        return_type: Literal["dataframe", "names", "idce", "idca", "dataarray", None],
        start_case=None,
        stop_case=None,
        step_case=None,
    ):
        if return_type:
            idx = self.datatree.caseids()
            if idx is not None:
                idx = idx[start_case:stop_case:step_case]
            if return_type == "names":
                return pd.DataFrame(
                    data=arr,
                    columns=self.graph.standard_sort_names[: arr.shape[1]],
                    index=idx,
                )
            if return_type == "dataarray":
                return DataArray(
                    arr,
                    coords={
                        self.datatree.CASEID: idx,
                        "nodeid": np.asarray(self.graph.standard_sort),
                        "node_name": DataArray(
                            np.asarray(self.graph.standard_sort_names), dims="nodeid"
                        ),
                    },
                    dims=(self.datatree.CASEID, "nodeid"),
                )
            result = pd.DataFrame(
                data=arr,
                columns=self.graph.standard_sort[: arr.shape[1]],
                index=idx,
            )
            if return_type == "idce":
                raise NotImplementedError
            elif return_type == "idca":
                return result.stack()
            else:
                return result
        return arr

    def probability(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        return_dataframe=False,
        include_nests=False,
    ):
        result_arrays, penalty = self._loglike_runner(
            x,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
            return_probability=True,
        )
        pr = result_arrays.probability
        if not include_nests:
            pr = pr[:, : self.graph.n_elementals()]
        return self._wrap_as_dataframe(
            pr,
            return_dataframe,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
        )

    def utility(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        return_format=None,
    ):
        """
        Compute values for the utility function contained in the model.

        Parameters
        ----------
        x : array-like or dict, optional
            New values to set for the parameters before evaluating
            the log likelihood.  If given as array-like, the array must
            be a vector with length equal to the length of the
            parameter frame, and the given vector will replace
            the current values.  If given as a dictionary,
            the dictionary is used to update the parameters.
        start_case : int, default 0
            The first case to include in the log likelihood computation.
            To include all cases, start from 0 (the default).
        stop_case : int, default -1
            One past the last case to include in the log likelihood
            computation.  This is processed as usual for Python slicing
            and iterating, and negative values count backward from the
            end.  To include all cases, end at -1 (the default).
        step_case : int, default 1
            The step size of the case iterator to use in likelihood
            calculation.  This is processed as usual for Python slicing
            and iterating.  To include all cases, step by 1 (the default).
        return_format : {None, 'idco', 'dataarray'}, default None
            Return the result in the indicated format.
              - 'idco' gives a pandas DataFrame indexed by cases and with
                alternatives as columns.
              - 'idca' gives a pandas Series with a two-level multi-index.
              - 'dataarray' gives a two-dimension larch DataArray.

        Returns
        -------
        array or DataFrame or DataArray
        """
        result_arrays, penalty = self._loglike_runner(
            x,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
            only_utility=2,
        )
        return self._wrap_as_dataframe(
            result_arrays.utility,
            return_format,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
        )

    def quantity(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        return_type: Literal[
            "dataframe", "names", "idce", "idca", "dataarray", None
        ] = None,
    ):
        result_arrays, penalty = self._loglike_runner(
            x,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
            only_utility=3,
        )
        return self._wrap_as_dataframe(
            result_arrays.utility,
            return_type,
        )

    def logsums(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        arr=None,
    ):
        result_arrays, penalty = self._loglike_runner(
            x,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
            only_utility=2,
        )
        if arr is not None:
            arr[start_case:stop_case:step_case] = result_arrays.utility[:, -1]
            return arr
        return result_arrays.utility[:, -1]

    def loglike2(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        persist=0,
        return_series=True,
        probability_only=False,
    ):
        result_arrays, penalty = self._loglike_runner(
            x,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
            return_gradient=True,
        )
        result = dictx(
            ll=result_arrays.loglike.sum() * self.weight_normalization,
            dll=result_arrays.d_loglike.sum(0) * self.weight_normalization,
        )
        if start_case is None and stop_case is None and step_case is None:
            self._check_if_best(result.ll)
        if return_series:
            result["dll"] = pd.Series(
                result["dll"],
                index=self.pnames,
            )
        return result

    def loglike2_bhhh(
        self,
        x=None,
        *,
        return_series=False,
        start_case=None,
        stop_case=None,
        step_case=None,
        persist=0,
    ):
        result_arrays, penalty = self._loglike_runner(
            x,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
            return_gradient=True,
            return_bhhh=True,
        )
        result = dictx(
            ll=result_arrays.loglike.sum() * self.weight_normalization,
            dll=result_arrays.d_loglike.sum(0) * self.weight_normalization,
            bhhh=result_arrays.bhhh.sum(0) * self.weight_normalization,
        )
        from ..model.persist_flags import (
            PERSIST_D_LOGLIKE_CASEWISE,
            PERSIST_LOGLIKE_CASEWISE,
        )

        if persist & PERSIST_LOGLIKE_CASEWISE:
            result["ll_casewise"] = result_arrays.loglike * self.weight_normalization
        if persist & PERSIST_D_LOGLIKE_CASEWISE:
            result["dll_casewise"] = result_arrays.d_loglike * self.weight_normalization
        if start_case is None and stop_case is None and step_case is None:
            self._check_if_best(result.ll)
        if return_series:
            result["dll"] = pd.Series(
                result["dll"],
                index=self.pnames,
            )
            result["bhhh"] = pd.DataFrame(
                result["bhhh"], index=self.pnames, columns=self.pnames
            )
        result["penalty"] = penalty
        return result

    def d2_loglike(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
    ):
        if x is None:
            x = self.pvals.copy()
        from ..util.math import approx_fprime

        result = approx_fprime(
            x,
            lambda y: self.d_loglike(
                y,
                start_case=start_case,
                stop_case=stop_case,
                step_case=step_case,
            ),
        )
        # the approx_fprime function will leave a residual epsilon on the last
        # parameter, we need to clean that up and restore the parameters to their
        # original values
        self.pvals = x
        return result

    def neg_loglike(
        self,
        x=None,
        start_case=None,
        stop_case=None,
        step_case=None,
    ):
        result = self.loglike(
            x,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
        )
        return -result

    def logloss(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        check_if_best=True,
    ):
        result = self.loglike(
            x,
            start_case=start_case,
            stop_case=stop_case,
            step_case=step_case,
            check_if_best=check_if_best,
        )
        return -result / self.total_weight()

    def neg_d_loglike(self, x=None, start_case=0, stop_case=-1, step_case=1, **kwargs):
        result = self.d_loglike(
            x, start_case=start_case, stop_case=stop_case, step_case=step_case, **kwargs
        )
        return -np.asarray(result)

    def d_logloss(self, x=None, start_case=0, stop_case=-1, step_case=1, **kwargs):
        result = self.d_loglike(
            x, start_case=start_case, stop_case=stop_case, step_case=step_case, **kwargs
        )
        return -np.asarray(result) / self.total_weight()

    def jumpstart_bhhh(
        self,
        steplen=0.5,
        jumpstart=0,
        jumpstart_split=5,
        logger=None,
    ):
        """
        Jump start optimization.

        Parameters
        ----------
        steplen
        jumpstart
        jumpstart_split

        """
        if logger is None:

            class NoLogger:
                debug = lambda *x: None
                info = lambda *x: None

            logger = NoLogger()

        for _jump in range(jumpstart):
            j_pvals = self.pvals.copy()
            #
            # jump_breaks = list(
            #     range(0, n_cases, n_cases // jumpstart_split + (1 if n_cases % jumpstart_split else 0))
            # ) + [n_cases]
            #
            # for j0, j1 in zip(jump_breaks[:-1], jump_breaks[1:]):
            for j0 in range(jumpstart_split):
                result = self.loglike2_bhhh(
                    start_case=j0,
                    step_case=jumpstart_split,
                )
                current_dll = result.dll
                current_bhhh = result.bhhh
                bhhh_inv = self._free_slots_inverse_matrix(current_bhhh)
                direction = np.dot(current_dll, bhhh_inv)
                j_pvals += direction * steplen
                logger.debug(f"jump to {j_pvals}")
                self.set_values(j_pvals)

    def check_d_loglike(self, stylize=True, skip_zeros=False):
        """
        Check that the analytic and finite-difference gradients are approximately equal.

        Primarily used for debugging.

        Parameters
        ----------
        stylize : bool, default True
            See :ref:`check_gradient` for details.
        skip_zeros : bool, default False
            See :ref:`check_gradient` for details.

        Returns
        -------
        pd.DataFrame or Stylized DataFrame
        """
        epsilon = np.sqrt(np.finfo(self.float_dtype).eps)
        from .gradient_check import check_gradient

        return check_gradient(
            self.loglike,
            self.d_loglike,
            self.pvals.copy(),
            names=list(self.pnames),
            stylize=stylize,
            skip_zeros=skip_zeros,
            epsilon=epsilon,
        )

    def __getstate__(self):
        attr = {}
        port = [
            "availability_any",
            "availability_ca_var",
            "availability_co_vars",
            "choice_any",
            "choice_ca_var",
            "choice_co_code",
            "choice_co_vars",
            "common_draws",
            "_compute_engine",
            "constraint_intensity",
            "constraint_sharpness",
            "constraints",
            # "dataflows",
            # 'dataset',
            # 'datatree',
            "float_dtype",
            "graph",
            "groupid",
            "logsum_parameter",
            "mixtures",
            "n_draws",
            "parameters",
            "prerolled_draws",
            "quantity_ca",
            "quantity_scale",
            "rename_parameters",
            "seed",
            "title",
            "utility_ca",
            "utility_co",
            "weight_co_var",
        ]
        for k in port:
            attr[k] = getattr(self, k)
        return attr

    def __setstate__(self, state):
        defer = {
            "constraints",
        }
        for check in [lambda i: i in defer, lambda i: i not in defer]:
            for k, v in state.items():
                if check(k):
                    continue
                if k == "parameters":
                    self._parameter_bucket.update_parameters(v)
                else:
                    try:
                        setattr(self, k, v)
                    except AttributeError as err:
                        raise AttributeError(f"{k}: {err}") from None

    def copy(self, datatree=True):
        dupe = type(self)()
        dupe.__setstate__(self.__getstate__())
        if datatree:
            dupe.datatree = self.datatree
        return dupe

    def remove_unused_parameters(self, verbose=True):
        """
        Remove parameters that are not used in the model.

        Parameters
        ----------
        verbose : bool, default True
            Generate log messages about how many parameters were dropped.
        """
        old_param_names = self.pnames
        old_parameters = self.parameters
        # clear existing parameters
        self._parameter_bucket._params = old_parameters.reindex(
            {self._parameter_bucket.index_name: []}
        )
        self.unmangle(True)
        new_param_names = self.pnames
        self._parameter_bucket._params = old_parameters.reindex(
            {self._parameter_bucket.index_name: new_param_names}
        )
        if verbose:
            dropped_params = [p for p in old_param_names if p not in new_param_names]
            if len(dropped_params) == 0:
                pass
            elif len(dropped_params) < 4:
                logging.getLogger("Larch").warning(
                    f"dropped {len(dropped_params)} parameters: {', '.join(dropped_params)}"
                )
            else:
                logging.getLogger("Larch").warning(
                    f"dropped {len(dropped_params)} parameters including: "
                    f"{', '.join(dropped_params[:3])}"
                )

    @property
    def n_cases(self) -> int:
        """The number of cases in the attached data."""
        data_as_possible = self.data_as_possible
        if data_as_possible is None:
            raise MissingDataError("no data are set")
        return data_as_possible.n_cases

    def total_weight(self) -> float:
        """
        Compute the total weight of cases in the loaded data.

        Returns
        -------
        float
        """
        if self._data_arrays is not None:
            if self._data_arrays.ch.ndim - 1 == self._data_arrays.wt.ndim:
                return (self._data_arrays.ch[:, -1] * self._data_arrays.wt).sum()
            elif self._data_arrays.ch.ndim - 2 == self._data_arrays.wt.ndim:
                return (
                    self._data_arrays.ch[:, -1]
                    * np.expand_dims(self._data_arrays.wt, -1)
                ).sum()
            elif self._data_arrays.ch.ndim - 3 == self._data_arrays.wt.ndim:
                return (
                    self._data_arrays.ch[:, -1]
                    * np.expand_dims(self._data_arrays.wt, (-1, -2))
                ).sum()
            else:
                raise ValueError
        if self.use_streaming:
            return self.streaming.total_weight()
        raise MissingDataError("no data_arrays are set")

    @property
    def dataset(self) -> Dataset | None:
        """xarray.Dataset : Data arrays as loaded for model computation."""
        super().unmangle()
        if self._dataset is None:
            self.reflow_data_arrays()
        try:
            return self._dataset
        except AttributeError:
            return None

    @dataset.setter
    def dataset(self, dataset: Dataset):
        if dataset is self._dataset:
            return
        if isinstance(dataset, Dataset):
            self._dataset = dataset
            self._data_arrays = None
        else:
            raise TypeError(f"dataset must be Dataset not {type(dataset)}")

    @dataset.deleter
    def dataset(self):
        self._dataset = None
        self._data_arrays = None

    @property
    def data_as_loaded(self):
        if self.dataset is not None:
            return self.dataset
        return None

    @property
    def data_as_possible(self):
        if self._dataset is not None:
            return self._dataset.dc
        if self.datatree is not None:
            return self.datatree
        return None

    @property
    def float_dtype(self):
        try:
            return self._float_dtype
        except AttributeError:
            return None

    @float_dtype.setter
    def float_dtype(self, float_dtype):
        if self.float_dtype != float_dtype:
            self.mangle()
        self._float_dtype = float_dtype

    def choice_avail_summary(self):
        """
        Generate a summary of choice and availability statistics.

        Returns
        -------
        pandas.DataFrame
        """
        from ..dataset.choice_avail_reporting import choice_avail_summary

        self.unmangle()
        graph = self.graph
        if self.use_streaming:
            return self.streaming.choice_avail_summary()
            # from .numba_stream import init_choice_avail_summary_streamer
            # choice_avail_summary_streamer = init_choice_avail_summary_streamer(self)
            # total_ch, total_av, total_wt = choice_avail_summary_streamer(
            #     self.n_cases,
            #     self._fixed_arrays.edge_slots,
            #     self._fixed_arrays.mu_slot,
            #     self.datatree.n_alts,
            # )
            # od = {}
            # idx = pd.Index(graph.standard_sort, name="altid")
            # od["name"] = pd.Series(graph.standard_sort_names, index=idx)
            # od["chosen"] = pd.Series(total_ch, index=idx)
            # od["available"] = pd.Series(total_av, index=idx)
            # result = pd.DataFrame(od,  index = idx)
            # from ..dataset.choice_avail_reporting import clean_summary
            # return clean_summary(result, root_id=graph.root_id)
        else:
            return choice_avail_summary(
                self.dataset,
                graph,
                self.availability_co_vars,
            )

    def maximize_loglike(
        self,
        *args,
        **kwargs,
    ) -> dictx:
        """
        Maximize the log likelihood.

        Parameters
        ----------
        method : str, optional
            The optimization method to use.  See scipy.optimize for
            most possibilities, or use 'BHHH'. Defaults to SLSQP if
            there are any constraints or finite parameter bounds,
            otherwise defaults to BHHH.
        quiet : bool, default False
            Whether to suppress the dashboard.
        options : dict, optional
            These options are passed through to the `scipy.optimize.minimize`
            function.
        maxiter : int, optional
            Maximum number of iterations.  This argument is just added to
            `options` for most methods.

        Returns
        -------
        larch.util.dictx
            A dictionary of results, including final log likelihood,
            elapsed time, and other statistics.  The exact items
            included in output will vary by estimation method.

        """
        from .optimization import maximize_loglike

        return maximize_loglike(self, *args, **kwargs)

    def _get_bounds_constraints(self, binding_tol=1e-4):
        """
        Convert bounds to parametric constraints on the model.

        Parameters
        ----------
        binding_tol : Number or Mapping
            The binding tolerance to use, which determines
            whether a constraint is considered active or not.

        Returns
        -------
        list
            A list of constraints.
        """
        from collections.abc import Mapping
        from numbers import Number

        from .constraints import FixedBound

        if isinstance(binding_tol, Number):
            default_binding_tol = binding_tol
        else:
            default_binding_tol = 1e-4
        if not isinstance(binding_tol, Mapping):
            binding_tol = {}
        get_bind_tol = lambda x: binding_tol.get(x, default_binding_tol)
        bounds = []
        self.unmangle()
        for pname, phold, pmin, pmax in zip(
            self.pnames,
            self.pholdfast,
            self.pminimum,
            self.pmaximum,
        ):
            if phold:
                # don't create bounds constraints on holdfast parameters
                continue
            if pmin != -np.inf or pmax != np.inf:
                bounds.append(
                    FixedBound(
                        pname, pmin, pmax, model=self, binding_tol=get_bind_tol(pname)
                    )
                )
        return bounds

    def _get_constraints(self, method):
        if method.lower() in ("slsqp", "cobyla"):
            constraint_dicts = []
            for c in self.constraints:
                constraint_dicts.extend(c.as_constraint_dicts())
            return constraint_dicts
        if method.lower() in ("trust-constr"):
            constraints = []
            for c in self.constraints:
                constraints.extend(c.as_linear_constraints())
            return constraints
        return ()

    def check_for_overspecification(self, pvals=None):
        """
        Check model for possible over-specification.

        Parameters
        ----------
        pvals : array-like, optional
            The parameter values to use in the calculation.  If not
            given, the current parameter values are used.

        Returns
        -------
        list of tuples
            A list of possible overspecification problems in the model.  Each problem
            is a tuple containing the eigenvalue, the indices of the non-zero elements
            in the eigenvector, and the eigenvector itself.
        """
        if pvals is None:
            pvals = self.pvals
        locks = np.asarray(self.pholdfast.astype(bool))
        if self.compute_engine == "jax":
            _se, hess, _inv_hess = self.jax_param_cov(pvals)
        else:
            hess = -self.d2_loglike(pvals)
        hess = np.asarray(hess).copy()
        hess[locks, :] = 0
        hess[:, locks] = 0

        overspec = compute_possible_overspecification(hess, self.pholdfast)
        if overspec:
            possible_overspecification = []
            msg = "Model is possibly over-specified (hessian is nearly singular)."
            msg += "\nLook for problems in these parameters or groups of parameters:"
            for eigval, ox, eigenvec in overspec:
                if eigval == "LinAlgError":
                    possible_overspecification.append((eigval, [ox], [""]))
                else:
                    paramset = list(np.asarray(self.pnames)[ox])
                    possible_overspecification.append((eigval, paramset, eigenvec[ox]))
                    msg += f"\n- Eigenvalue: {eigval}"
                    max_len_param = max(len(p) for p in paramset)
                    for p, z in zip(paramset, eigenvec[ox]):
                        msg += f"\n    {p:{max_len_param}s}: {z}"
            self._possible_overspecification = possible_overspecification
            warnings.warn(
                msg,
                category=PossibleOverspecification,
                stacklevel=2,
            )
            return possible_overspecification

    def calculate_parameter_covariance(self, pvals=None, *, robust=False):
        """
        Calculate the parameter covariance matrix.

        Parameters
        ----------
        pvals : array-like, optional
            The parameter values to use in the calculation.  If not
            given, the current parameter values are used.
        robust : bool, default False
            Whether to calculate the robust covariance matrix.

        Returns
        -------
        se : array
            The standard errors of the parameter estimates.
        hess : array
            The Hessian matrix.
        ihess : array
            The inverse of the Hessian matrix.
        """
        if pvals is None:
            pvals = self.pvals
        locks = np.asarray(self.pholdfast.astype(bool))
        if self.compute_engine == "jax":
            se, hess, ihess = self.jax_param_cov(pvals)
        else:
            hess = -self.d2_loglike(pvals)
            if self.parameters["holdfast"].sum():
                free = self.pholdfast == 0
                hess_ = hess[free][:, free]
                try:
                    ihess_ = np.linalg.inv(hess_)
                except np.linalg.LinAlgError:
                    ihess_ = np.linalg.pinv(hess_)
                ihess = _arr_inflate(ihess_, locks)
            else:
                try:
                    ihess = np.linalg.inv(hess)
                except np.linalg.LinAlgError:
                    ihess = np.linalg.pinv(hess)
            se = _safe_sqrt(ihess.diagonal())
            self.pstderr = se
        hess = np.asarray(hess).copy()
        hess[locks, :] = 0
        hess[:, locks] = 0
        ihess = np.asarray(ihess).copy()
        ihess[locks, :] = 0
        ihess[:, locks] = 0
        self.add_parameter_array("hess", hess)
        self.add_parameter_array("ihess", ihess)

        overspec = compute_possible_overspecification(hess, self.pholdfast)
        if overspec:
            warnings.warn(
                "Model is possibly over-specified (hessian is nearly singular).",
                category=PossibleOverspecification,
                stacklevel=2,
            )
            possible_overspecification = []
            for eigval, ox, eigenvec in overspec:
                if eigval == "LinAlgError":
                    possible_overspecification.append((eigval, [ox], [""]))
                else:
                    paramset = list(np.asarray(self.pnames)[ox])
                    possible_overspecification.append((eigval, paramset, eigenvec[ox]))
            self._possible_overspecification = possible_overspecification

        # constrained covariance
        if self.constraints:
            constraints = list(self.constraints)
        else:
            constraints = []
        try:
            constraints.extend(self._get_bounds_constraints())
        except Exception:
            pass

        if constraints:
            binding_constraints = list()
            self.add_parameter_array("unconstrained_std_err", self.pstderr)
            self.add_parameter_array("unconstrained_covariance_matrix", ihess)

            s = np.asarray(ihess)
            pvals = self.pvals
            for c in constraints:
                if np.absolute(c.fun(pvals)) < c.binding_tol:
                    binding_constraints.append(c)
                    b = c.jac(self.pf.value)
                    den = b @ s @ b
                    if den != 0:
                        s = s - (1 / den) * s @ b.reshape(-1, 1) @ b.reshape(1, -1) @ s
            self.add_parameter_array("covariance_matrix", s)
            self.pstderr = _safe_sqrt(s.diagonal())

            # Fix numerical issues on some constraints, add constrained notes
            if binding_constraints or any(self.pholdfast != 0):
                notes = {}
                for c in binding_constraints:
                    pa = c.get_parameters()
                    for p in pa:
                        # if self.pf.loc[p, 't_stat'] > 1e5:
                        #     self.pf.loc[p, 't_stat'] = np.inf
                        #     self.pf.loc[p, 'std_err'] = np.nan
                        # if self.pf.loc[p, 't_stat'] < -1e5:
                        #     self.pf.loc[p, 't_stat'] = -np.inf
                        #     self.pf.loc[p, 'std_err'] = np.nan
                        n = notes.get(p, [])
                        n.append(c.get_binding_note(pvals))
                        notes[p] = n
                constrained_note = (
                    pd.Series({k: "\n".join(v) for k, v in notes.items()}, dtype=object)
                    .reindex(self.pnames)
                    .fillna("")
                )
                constrained_note[self.pholdfast != 0] = "fixed value"
                self.add_parameter_array("constrained", constrained_note)

        if robust:
            self.robust_covariance()
            se = self.parameters["robust_std_err"]

        return se, hess, ihess

    def histogram_on_idca_variable(
        self,
        x,
        **kwargs,
    ):
        from ..util.figures import histogram_on_idca_variable

        return histogram_on_idca_variable(
            x,
            pr=self.probability(),  # array-like
            ds=self.dataset,
            dt=self.datatree,
            **kwargs,
        )

    def utility_breakdown(
        self,
        altid: int,
        *,
        caseid: int | None = None,
        caseindex: int | None = None,
    ) -> pd.DataFrame:
        """
        Compute the utility breakdown for a given case and alternative.

        One and only one of `caseid` or `caseindex` must be given as a keyword
        argument, to indicate which case to compute the utility breakdown for.

        Parameters
        ----------
        altid : int
            The alternative ID on which to compute the utility breakdown.
        caseid : int, optional
            The case ID on which to compute the utility breakdown.
        caseindex : int, optional
            The case index on which to compute the utility breakdown.

        Returns
        -------
        pandas.DataFrame
        """
        if caseid is None and caseindex is None:
            raise ValueError("either caseid or caseindex must be given")
        if caseid is not None and caseindex is not None:
            raise ValueError("only one of caseid or caseindex must be given")
        if caseid is not None:
            caseindex = self.dataset.indexes[self.dataset.dc.CASEID].get_loc(caseid)
        df = {}
        df["co"] = pd.DataFrame(
            [
                (
                    k.param,
                    k.data,
                    float(self.dataset["co"][caseindex].sel(var_co=k.data)),
                    float(self.get_value(k.param)),
                )
                for k in self.utility_co[altid]
            ],
            columns=["parameter_name", "data_expr", "data_value", "parameter_value"],
        )
        df["ca"] = pd.DataFrame(
            [
                (
                    k.param,
                    k.data,
                    float(
                        self.dataset["ca"][caseindex].sel(
                            **{self.dataset.dc.ALTID: altid, "var_ca": k.data}
                        )
                    ),
                    float(self.get_value(k.param)),
                )
                for k in self.utility_ca
            ],
            columns=["parameter_name", "data_expr", "data_value", "parameter_value"],
        )
        result = pd.concat(df, keys=["co", "ca"], names=["utility_type"])
        result["partial_utility"] = result["data_value"] * result["parameter_value"]
        return result.set_index(["utility_type", "parameter_name", "data_expr"])

    def release_memory(self):
        """Release memory-intensive data structures."""
        self._fixed_arrays = None
        self._data_arrays = None
        self._dataset = None
        self.work_arrays = None


@njit(cache=True)
def _arr_inflate(arr, locks):
    s = locks.size
    z = np.zeros((s, s), dtype=arr.dtype)
    i_ = 0
    for i in range(s):
        if not locks[i]:
            j_ = 0
            for j in range(s):
                if not locks[j]:
                    z[i, j] = arr[i_, j_]
                    j_ += 1
            i_ += 1
    return z


def _safe_sqrt(x):
    result = np.sqrt(np.clip(x, 0, np.inf))
    result[x < 0] = np.nan
    return result



from __future__ import annotations

import ast
import base64
import hashlib
import logging
import re
import time
from collections.abc import Hashable, Iterable, Mapping, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd
import pyarrow as pa
import xarray as xr
from xarray import DataArray, Dataset

from .accessors import register_dataset_method
from .aster import extract_all_name_tokens
from .categorical import _Categorical  # noqa
from .shared_memory import si_units
from .table import Table

if TYPE_CHECKING:
    import openmatrix


logger = logging.getLogger("sharrow")

well_known_names = {
    "nb",
    "np",
    "pd",
    "xr",
    "pa",
    "log",
    "exp",
    "log1p",
    "expm1",
    "max",
    "min",
    "piece",
    "hard_sigmoid",
    "transpose_leading",
    "clip",
}


def one_based(n):
    return pd.RangeIndex(1, n + 1)


def zero_based(n):
    return pd.RangeIndex(0, n)


def clean(s):
    """
    Convert any string into a similar python identifier.

    If any modification of the string is made, or if the string
    is longer than 120 characters, it is truncated and a hash of the
    original string is added to the end, to ensure every
    string maps to a unique cleaned name.

    Parameters
    ----------
    s : str

    Returns
    -------
    cleaned : str
    """
    if not isinstance(s, str):
        s = f"{type(s)}-{s}"
    cleaned = re.sub(r"\W|^(?=\d)", "_", s)
    if cleaned != s or len(cleaned) > 120:
        # digest size 15 creates a 24 character base32 string
        h = base64.b32encode(
            hashlib.blake2b(s.encode(), digest_size=15).digest()
        ).decode()
        cleaned = f"{cleaned[:90]}_{h}"
    return cleaned


def construct(source):
    """
    Create Datasets from various similar objects.

    Parameters
    ----------
    source : pandas.DataFrame, pyarrow.Table, xarray.Dataset, or Sequence[str]
        The source from which to create a Dataset.  DataFrame and Table objects
        are converted to Datasets that have one dimension (the rows) and
        separate variables for each of the columns.  A list of strings
        creates a dataset with those named empty variables.

    Returns
    -------
    Dataset
    """
    if isinstance(source, pd.DataFrame):
        source = dataset_from_dataframe_fast(source)  # xarray default can be slow
    elif isinstance(source, (Table, pa.Table)):
        source = from_table(source)
    elif isinstance(source, xr.Dataset):
        pass  # don't do the superclass things
    elif isinstance(source, Sequence) and all(isinstance(i, str) for i in source):
        source = from_table(pa.table({i: [] for i in source}))
    else:
        source = xr.Dataset(source)
    return source


def dataset_from_dataframe_fast(
    dataframe: pd.DataFrame,
    sparse: bool = False,
    preserve_cat: bool = True,
) -> Dataset:
    """Convert a pandas.DataFrame into an xarray.Dataset.

    Each column will be converted into an independent variable in the
    Dataset. If the dataframe's index is a MultiIndex, it will be expanded
    into a tensor product of one-dimensional indices (filling in missing
    values with NaN). This method will produce a Dataset very similar to
    that on which the 'to_dataframe' method was called, except with
    possibly redundant dimensions (since all dataset variables will have
    the same dimensionality)

    Parameters
    ----------
    dataframe : DataFrame
        DataFrame from which to copy data and indices.
    sparse : bool, default: False
        If true, create a sparse arrays instead of dense numpy arrays. This
        can potentially save a large amount of memory if the DataFrame has
        a MultiIndex. Requires the sparse package (sparse.pydata.org).
    preserve_cat : bool, default True
        If true, preserve encoding of categorical columns.  Xarray lacks an
        official implementation of a categorical datatype, so sharrow's
        dictionary-based digital encoding is applied instead. Note that in
        native xarray usage, the resulting variable will look like integer
        values instead of the category values.  The `dataset.cat` accessor
        can be used to interact with the categorical data.

    Returns
    -------
    New Dataset.

    See Also
    --------
    xarray.DataArray.from_series
    pandas.DataFrame.to_xarray
    """
    # this is much faster than the default xarray version when not
    # using a MultiIndex.

    if isinstance(dataframe.index, pd.MultiIndex) or sparse:
        return Dataset.from_dataframe(dataframe, sparse)

    if not dataframe.columns.is_unique:
        # if the dataframe has non-unique column names, but all the duplicate
        # names contain the same data, we can recover safely by dropping the
        # duplicates, otherwise throw an error.
        cannot_fix = False
        dupe_columns = dataframe.columns.duplicated()
        dupe_column_names = dataframe.columns[dupe_columns]
        for j in dupe_column_names:
            subframe = dataframe[j]
            ref_col = subframe.iloc[:, 0]
            for k in range(1, len(subframe.columns)):
                if not ref_col.equals(subframe.iloc[:, k]):
                    cannot_fix = True
                    break
                if cannot_fix:
                    break
        dupe_column_names = [f"- {i}" for i in dupe_column_names]
        logger.error(
            "DataFrame has non-unique columns\n" + "\n".join(dupe_column_names)
        )
        if cannot_fix:
            raise ValueError("cannot convert DataFrame with non-unique columns")
        else:
            dataframe = dataframe.loc[:, ~dupe_columns]

    if isinstance(dataframe.index, pd.CategoricalIndex):
        idx = dataframe.index.remove_unused_categories()
    else:
        idx = dataframe.index

    index_name = idx.name if idx.name is not None else "index"
    # Cast to a NumPy array first, in case the Series is a pandas Extension
    # array (which doesn't have a valid NumPy dtype)
    arrays = {}
    for name in dataframe.columns:
        if name != index_name:
            if dataframe[name].dtype == "category" and preserve_cat:
                cat = dataframe[name].cat
                categories = np.asarray(cat.categories)
                if categories.dtype.kind == "O":
                    categories = categories.astype(str)
                arrays[name] = (
                    [index_name],
                    np.asarray(cat.codes),
                    {
                        "digital_encoding": {
                            "dictionary": categories,
                            "ordered": cat.ordered,
                        }
                    },
                )
            else:
                arrays[name] = ([index_name], np.asarray(dataframe[name].values))
    return Dataset(arrays, coords={index_name: (index_name, dataframe.index.values)})


def from_table(
    tbl,
    index_name="index",
    index=None,
):
    """
    Convert a pyarrow.Table into an xarray.Dataset.

    Parameters
    ----------
    tbl : Table
        Table from which to use data and indices.
    index_name : str, default 'index'
        This name will be given to the default dimension index, if
        none is given.  Ignored if `index` is given explicitly and
        it already has a name.
    index : Index-like, optional
        Use this index instead of a default RangeIndex.

    Returns
    -------
    New Dataset.
    """
    if len(set(tbl.column_names)) != len(tbl.column_names):
        raise ValueError("cannot convert Table with non-unique columns")
    if index is None:
        index = pd.RangeIndex(len(tbl), name=index_name)
    else:
        if len(index) != len(tbl):
            raise ValueError(
                f"length of index ({len(index)}) does not match length of table ({len(tbl)})"
            )
    if isinstance(index, pd.MultiIndex) and not index.is_unique:
        raise ValueError(
            "cannot attach a non-unique MultiIndex and convert into xarray"
        )
    arrays = []
    metadata = {}
    for n in range(len(tbl.column_names)):
        c = tbl.column(n)
        if isinstance(c.type, pa.DictionaryType):
            cc = c.combine_chunks()
            arrays.append((tbl.column_names[n], np.asarray(cc.indices)))
            metadata[tbl.column_names[n]] = {
                "digital_encoding": {
                    "dictionary": cc.dictionary,
                    "ordered": cc.type.ordered,
                }
            }
        else:
            arrays.append((tbl.column_names[n], np.asarray(c)))
    result = xr.Dataset()
    if isinstance(index, pd.MultiIndex):
        dims = tuple(
            name if name is not None else f"level_{n}"
            for n, name in enumerate(index.names)
        )
        for dim, lev in zip(dims, index.levels):
            result[dim] = (dim, lev)
    else:
        try:
            if index.name is not None:
                index_name = index.name
        except AttributeError:
            pass
        dims = (index_name,)
        result[index_name] = (dims, index)

    result._set_numpy_data_from_dataframe(index, arrays, dims)
    for k, v in metadata.items():
        result[k].attrs.update(v)
    return result


def from_omx(
    omx: openmatrix.File,
    index_names=("otaz", "dtaz"),
    indexes="one-based",
    renames=None,
):
    """
    Create a Dataset from an OMX file.

    Parameters
    ----------
    omx : openmatrix.File or larch.OMX
        An OMX-format file, opened for reading.
    index_names : tuple, default ("otaz", "dtaz")
        Should be a tuple of length 3, giving the names of the three
        dimensions.  The first two names are the native dimensions from
        the open matrix file, the last is the name of the implicit
        dimension that is created by parsing array names.
    indexes : str or tuple[str], optional
        The name of a 'lookup' in the OMX file, which will be used to
        populate the coordinates for the two native dimensions.  Or,
        specify "one-based" or "zero-based" to assume sequential and
        consecutive numbering starting with 1 or 0 respectively. For
        non-square OMX data, this must be given as a tuple, relating
        indexes as above for each dimension of `index_names`.
    renames : Mapping or Collection, optional
        Limit the import only to these data elements.  If given as a
        mapping, the keys will be the names of variables in the resulting
        dataset, and the values give the names of data matrix tables in the
        OMX file.  If given as a list or other non-mapping collection,
        elements are not renamed but only elements in the collection are
        included.

    Returns
    -------
    Dataset
    """
    # handle both larch.OMX and openmatrix.open_file versions
    if "lar" in type(omx).__module__:
        omx_data = omx.data
        omx_lookup = omx.lookup
        omx_shape = omx.shape
    else:
        omx_data = omx.root["data"]
        omx_lookup = omx.root["lookup"]
        omx_shape = omx.shape()

    arrays = {}
    if renames is None:
        for k in omx_data._v_children:
            arrays[k] = omx_data[k][:]
    elif isinstance(renames, dict):
        for new_k, old_k in renames.items():
            arrays[new_k] = omx_data[old_k][:]
    else:
        for k in renames:
            arrays[k] = omx_data[k][:]
    d = {
        "dims": index_names,
        "data_vars": {k: {"dims": index_names, "data": arrays[k]} for k in arrays},
    }
    if indexes == "one-based":
        indexes = {
            index_names[0]: one_based(omx_shape[0]),
            index_names[1]: one_based(omx_shape[1]),
        }
    elif indexes == "zero-based":
        indexes = {
            index_names[0]: zero_based(omx_shape[0]),
            index_names[1]: zero_based(omx_shape[1]),
        }
    elif isinstance(indexes, str):
        if indexes in omx_lookup:
            if omx_shape[0] != omx_shape[1]:
                raise ValueError("singleton arbitrary coordinates on non-square arrays")
            ixs = np.asarray(omx_lookup[indexes])
            indexes = {
                index_names[0]: ixs,
                index_names[1]: ixs,
            }
        else:
            raise KeyError(f"{indexes} not found in OMX lookups")
    elif isinstance(indexes, tuple):
        indexes_ = {}
        for n, (name, i) in enumerate(zip(index_names, indexes)):
            if i == "one-based":
                indexes_[name] = one_based(omx_shape[n])
            elif i == "zero-based":
                indexes_[name] = zero_based(omx_shape[n])
            elif isinstance(i, str):
                if i in omx_lookup:
                    indexes_[name] = np.asarray(omx_lookup[i])
                else:
                    raise KeyError(f"{i} not found in OMX lookups")
        indexes = indexes_
    if indexes is not None:
        d["coords"] = {
            index_name: {"dims": index_name, "data": index}
            for index_name, index in indexes.items()
        }
    return xr.Dataset.from_dict(d)


def _should_ignore(ignore, x):
    if ignore is not None:
        for i in ignore:
            if re.match(i, x):
                return True
    return False


def from_omx_3d(
    omx: openmatrix.File | str | Iterable[openmatrix.File | str],
    index_names=("otaz", "dtaz", "time_period"),
    indexes=None,
    *,
    time_periods=None,
    time_period_sep="__",
    max_float_precision=32,
    ignore=None,
):
    """
    Create a Dataset from an OMX file with an implicit third dimension.

    Parameters
    ----------
    omx : openmatrix.File or larch.OMX
        An OMX-format file, opened for reading.
    index_names : tuple, default ("otaz", "dtaz", "time_period")
        Should be a tuple of length 3, giving the names of the three
        dimensions.  The first two names are the native dimensions from
        the open matrix file, the last is the name of the implicit
        dimension that is created by parsing array names.
    indexes : str, optional
        The name of a 'lookup' in the OMX file, which will be used to
        populate the coordinates for the two native dimensions.  Or,
        specify "one-based" or "zero-based" to assume sequential and
        consecutive numbering starting with 1 or 0 respectively.
    time_periods : list-like, required keyword argument
        A list of index values from which the third dimension is constructed
        for all variables with a third dimension.
    time_period_sep : str, default "__" (double underscore)
        The presence of this separator within the name of any table in the
        OMX file indicates that table is to be considered a page in a
        three dimensional variable.  The portion of the name preceding the
        first instance of this separator is the name of the resulting
        variable, and the portion of the name after the first instance of
        this separator is the label of the position for this page, which
        should appear in `time_periods`.
    max_float_precision : int, default 32
        When loading, reduce all floats in the OMX file to this level of
        precision, generally to save memory if they were stored as double
        precision but that level of detail is unneeded in the present
        application.
    ignore : list-like, optional
        A list of regular expressions that will be used to filter out
        variables from the dataset.  If any of the regular expressions
        match the name of a variable, that variable will not be included
        in the loaded dataset. This is useful for excluding variables that
        are not needed in the current application.

    Returns
    -------
    Dataset
    """
    if not isinstance(omx, (list, tuple)):
        omx = [omx]

    use_file_handles = []
    opened_file_handles = []
    for filename in omx:
        if isinstance(filename, str):
            import openmatrix

            h = openmatrix.open_file(filename)
            opened_file_handles.append(h)
            use_file_handles.append(h)
        else:
            use_file_handles.append(filename)
    omx = use_file_handles

    try:
        # handle both larch.OMX and openmatrix.open_file versions
        if "larch" in type(omx[0]).__module__:
            omx_shape = omx[0].shape
            omx_lookup = omx[0].lookup
        else:
            omx_shape = omx[0].shape()
            omx_lookup = omx[0].root["lookup"]
        omx_data = []
        omx_data_map = {}
        for n, i in enumerate(omx):
            if "larch" in type(i).__module__:
                omx_data.append(i.data)
                for k in i.data._v_children:
                    omx_data_map[k] = n
            else:
                omx_data.append(i.root["data"])
                for k in i.root["data"]._v_children:
                    omx_data_map[k] = n

        import dask.array

        data_names = list(omx_data_map.keys())
        if ignore is not None:
            if isinstance(ignore, str):
                ignore = [ignore]
            data_names = [i for i in data_names if not _should_ignore(ignore, i)]
        n1, n2 = omx_shape
        if indexes is None:
            # default reads mapping if only one lookup is included, otherwise one-based
            if len(omx_lookup._v_children) == 1:
                ranger = None
                indexes = list(omx_lookup._v_children)[0]
            else:
                ranger = one_based
        elif indexes == "one-based":
            ranger = one_based
        elif indexes == "zero-based":
            ranger = zero_based
        elif indexes in set(omx_lookup._v_children):
            ranger = None
        else:
            raise NotImplementedError(
                "only one-based, zero-based, and named indexes are implemented"
            )
        if ranger is not None:
            r1 = ranger(n1)
            r2 = ranger(n2)
        else:
            r1 = r2 = pd.Index(omx_lookup[indexes])

        if time_periods is None:
            raise ValueError("must give time periods explicitly")

        time_periods_map = {t: n for n, t in enumerate(time_periods)}

        pending_3d = {}
        content = {}

        for k in data_names:
            if time_period_sep in k:
                base_k, time_k = k.split(time_period_sep, 1)
                if base_k not in pending_3d:
                    pending_3d[base_k] = [None] * len(time_periods)
                pending_3d[base_k][time_periods_map[time_k]] = dask.array.from_array(
                    omx_data[omx_data_map[k]][k]
                )
            else:
                content[k] = xr.DataArray(
                    dask.array.from_array(omx_data[omx_data_map[k]][k]),
                    dims=index_names[:2],
                    coords={
                        index_names[0]: r1,
                        index_names[1]: r2,
                    },
                )
        for base_k, darrs in pending_3d.items():
            # find a prototype array
            prototype = None
            for i in darrs:
                prototype = i
                if prototype is not None:
                    break
            if prototype is None:
                raise ValueError("no prototype")
            darrs_ = [
                (i if i is not None else dask.array.zeros_like(prototype))
                for i in darrs
            ]
            content[base_k] = xr.DataArray(
                dask.array.stack(darrs_, axis=-1),
                dims=index_names,
                coords={
                    index_names[0]: r1,
                    index_names[1]: r2,
                    index_names[2]: time_periods,
                },
            )
        for i in content:
            if np.issubdtype(content[i].dtype, np.floating):
                if content[i].dtype.itemsize > max_float_precision / 8:
                    content[i] = content[i].astype(f"float{max_float_precision}")
        return xr.Dataset(content)
    finally:
        for h in opened_file_handles:
            h.close()


def reload_from_omx_3d(
    dataset: xr.Dataset,
    omx: Iterable[str],
    *,
    time_period_sep="__",
    ignore=None,
) -> None:
    """
    Reload the content of a dataset from OMX files.

    This loads the data from the OMX files into the dataset, replacing
    the existing data in the dataset.  The dataset must have been created
    by `from_omx_3d` or a similar function. Note that `from_omx_3d` will
    create a dataset backed by `dask.array` objects; this function allows for
    loading the data without going through dask, which may have poor performance
    on some platforms.

    Parameters
    ----------
    dataset : xr.Dataset
        The dataset to reload into.
    omx : Iterable[str]
        The list of OMX file names to load from.
    time_period_sep : str, default "__"
        The separator used to identify time periods in the dataset.
    ignore : list-like, optional
        A list of regular expressions that will be used to filter out
        variables from the dataset.  If any of the regular expressions
        match the name of a variable, that variable will not be included
        in the load process. This is useful for excluding variables that
        are not found in the target dataset.
    """
    if isinstance(ignore, str):
        ignore = [ignore]

    use_file_handles = []
    opened_file_handles = []
    for filename in omx:
        if isinstance(filename, (str, Path)):
            import openmatrix

            h = openmatrix.open_file(filename)
            opened_file_handles.append(h)
            use_file_handles.append(h)
        else:
            use_file_handles.append(filename)
    omx = use_file_handles

    bytes_loaded = 0

    try:
        t0 = time.time()
        for filename, f in zip(omx, use_file_handles):
            if isinstance(filename, str):
                logger.info(f"loading into dataset from {filename}")
            for data_name in f.root.data._v_children:
                if _should_ignore(ignore, data_name):
                    logger.info(f"ignoring {data_name}")
                    continue
                t1 = time.time()
                filters = f.root.data[data_name].filters
                filter_note = f"{filters.complib}/{filters.complevel}"

                if time_period_sep in data_name:
                    data_name_x, data_name_t = data_name.split(time_period_sep, 1)
                    if data_name_x not in dataset:
                        logger.info(
                            f"skipping {data_name} because {data_name_x} not in dataset"
                        )
                        continue
                    if len(dataset[data_name_x].dims) != 3:
                        raise ValueError(
                            f"dataset variable {data_name_x} has "
                            f"{len(dataset[data_name_x].dims)} dimensions, expected 3"
                        )
                    raw = dataset[data_name_x].sel(time_period=data_name_t).data
                    raw[:, :] = f.root.data[data_name][:, :]
                else:
                    if len(dataset[data_name].dims) != 2:
                        raise ValueError(
                            f"dataset variable {data_name} has "
                            f"{len(dataset[data_name].dims)} dimensions, expected 2"
                        )
                    raw = dataset[data_name].data
                    raw[:, :] = f.root.data[data_name][:, :]
                bytes_loaded += raw.nbytes
                logger.info(
                    f"loaded {data_name} ({filter_note}) to dataset "
                    f"in {time.time() - t1:.2f}s, {si_units(bytes_loaded)}"
                )
        logger.info(f"loading to dataset complete in {time.time() - t0:.2f}s")
    finally:
        for h in opened_file_handles:
            h.close()


def from_amx(
    amx,
    index_names=("otaz", "dtaz"),
    indexes="one-based",
    renames=None,
):
    arrays = {}
    if renames is None:
        for k in amx.list_matrices():
            arrays[k] = amx[k][:]
    elif isinstance(renames, dict):
        for new_k, old_k in renames.items():
            arrays[new_k] = amx[old_k]
    else:
        for k in renames:
            arrays[k] = amx[k]
    d = {
        "dims": index_names,
        "data_vars": {k: {"dims": index_names, "data": arrays[k]} for k in arrays},
    }
    if indexes == "one-based":
        indexes = {index_names[i]: "1" for i in range(len(index_names))}
    elif indexes == "zero-based":
        indexes = {index_names[i]: "0" for i in range(len(index_names))}
    if isinstance(indexes, (list, tuple)):
        indexes = dict(zip(index_names, indexes))
    if isinstance(indexes, dict):
        for n, i in enumerate(index_names):
            if indexes.get(i) == "1":
                indexes[i] = one_based(amx.shape[n])
            elif indexes.get(i) == "0":
                indexes[i] = zero_based(amx.shape[n])
    if indexes is not None:
        d["coords"] = {
            index_name: {"dims": index_name, "data": index}
            for index_name, index in indexes.items()
        }
    return xr.Dataset.from_dict(d)


def from_zarr(store, *args, **kwargs):
    """
    Load and decode a dataset from a Zarr store.

    The `store` object should be a valid store for a Zarr group. `store`
    variables must contain dimension metadata encoded in the
    `_ARRAY_DIMENSIONS` attribute.

    Parameters
    ----------
    store : MutableMapping or str
        A MutableMapping where a Zarr Group has been stored or a path to a
        directory in file system where a Zarr DirectoryStore has been stored.
    synchronizer : object, optional
        Array synchronizer provided to zarr
    group : str, optional
        Group path. (a.k.a. `path` in zarr terminology.)
    chunks : int or dict or tuple or {None, 'auto'}, optional
        Chunk sizes along each dimension, e.g., ``5`` or
        ``{'x': 5, 'y': 5}``. If `chunks='auto'`, dask chunks are created
        based on the variable's zarr chunks. If `chunks=None`, zarr array
        data will lazily convert to numpy arrays upon access. This accepts
        all the chunk specifications as Dask does.
    overwrite_encoded_chunks : bool, optional
        Whether to drop the zarr chunks encoded for each variable when a
        dataset is loaded with specified chunk sizes (default: False)
    decode_cf : bool, optional
        Whether to decode these variables, assuming they were saved according
        to CF conventions.
    mask_and_scale : bool, optional
        If True, replace array values equal to `_FillValue` with NA and scale
        values according to the formula `original_values * scale_factor +
        add_offset`, where `_FillValue`, `scale_factor` and `add_offset` are
        taken from variable attributes (if they exist).  If the `_FillValue` or
        `missing_value` attribute contains multiple values a warning will be
        issued and all array values matching one of the multiple values will
        be replaced by NA.
    decode_times : bool, optional
        If True, decode times encoded in the standard NetCDF datetime format
        into datetime objects. Otherwise, leave them encoded as numbers.
    concat_characters : bool, optional
        If True, concatenate along the last dimension of character arrays to
        form string arrays. Dimensions will only be concatenated over (and
        removed) if they have no corresponding variable and if they are only
        used as the last dimension of character arrays.
    decode_coords : bool, optional
        If True, decode the 'coordinates' attribute to identify coordinates in
        the resulting dataset.
    drop_variables : str or iterable, optional
        A variable or list of variables to exclude from being parsed from the
        dataset. This may be useful to drop variables with problems or
        inconsistent values.
    consolidated : bool, optional
        Whether to open the store using zarr's consolidated metadata
        capability. Only works for stores that have already been consolidated.
        By default (`consolidate=None`), attempts to read consolidated metadata,
        falling back to read non-consolidated metadata if that fails.
    chunk_store : MutableMapping, optional
        A separate Zarr store only for chunk data.
    storage_options : dict, optional
        Any additional parameters for the storage backend (ignored for local
        paths).
    decode_timedelta : bool, optional
        If True, decode variables and coordinates with time units in
        {'days', 'hours', 'minutes', 'seconds', 'milliseconds', 'microseconds'}
        into timedelta objects. If False, leave them encoded as numbers.
        If None (default), assume the same value of decode_time.
    use_cftime : bool, optional
        Only relevant if encoded dates come from a standard calendar
        (e.g. "gregorian", "proleptic_gregorian", "standard", or not
        specified).  If None (default), attempt to decode times to
        ``np.datetime64[ns]`` objects; if this is not possible, decode times to
        ``cftime.datetime`` objects. If True, always decode times to
        ``cftime.datetime`` objects, regardless of whether or not they can be
        represented using ``np.datetime64[ns]`` objects.  If False, always
        decode times to ``np.datetime64[ns]`` objects; if this is not possible
        raise an error.

    Returns
    -------
    dataset : Dataset
        The newly created dataset.

    References
    ----------
    http://zarr.readthedocs.io/
    """
    return xr.open_zarr(store, *args, **kwargs)


def from_zarr_with_attr(*args, **kwargs):
    obj = from_zarr(*args, **kwargs)
    for k in obj:
        attrs = {}
        for aname, avalue in obj[k].attrs.items():
            attrs[aname] = _from_evalable_string(avalue)
        obj[k] = obj[k].assign_attrs(attrs)
    attrs = {}
    for aname, avalue in obj.attrs.items():
        attrs[aname] = _from_evalable_string(avalue)
    obj = obj.assign_attrs(attrs)
    return obj


def coerce_to_range_index(idx):
    if isinstance(idx, pd.RangeIndex):
        return idx
    if isinstance(idx, (pd.Int64Index, pd.Float64Index, pd.UInt64Index)):
        if idx.is_monotonic_increasing and idx[-1] - idx[0] == idx.size - 1:
            return pd.RangeIndex(idx[0], idx[0] + idx.size)
    return idx


def is_dict_like(value: Any) -> bool:
    return hasattr(value, "keys") and hasattr(value, "__getitem__")


@xr.register_dataset_accessor("single_dim")
class _SingleDim:
    """Convenience accessor for single-dimension datasets."""

    __slots__ = ("dataset", "dim_name")

    def __init__(self, dataset: Dataset):
        self.dataset = dataset
        if len(self.dataset.dims) != 1:
            raise ValueError("single_dim implies a single dimension dataset")
        self.dim_name = self.dataset.dims.__iter__().__next__()

    @property
    def coords(self):
        return self.dataset.coords[self.dim_name]

    @property
    def index(self):
        return self.dataset.indexes[self.dim_name]

    @property
    def size(self):
        return self.dataset.dims[self.dim_name]

    def _to_pydict(self):
        columns = [k for k in self.dataset.variables if k != self.dim_name]
        data = []
        for k in columns:
            a = self.dataset._variables[k]
            if (
                "digital_encoding" in a.attrs
                and "dictionary" in a.attrs["digital_encoding"]
            ):
                de = a.attrs["digital_encoding"]
                data.append(
                    pd.Categorical.from_codes(
                        a.values,
                        de["dictionary"],
                        de.get("ordered"),
                    )
                )
            else:
                data.append(a.values)
        return dict(zip(columns, data))

    def to_pyarrow(self) -> pa.Table:
        columns = [k for k in self.dataset.variables if k != self.dim_name]
        data = []
        for k in columns:
            a = self.dataset._variables[k]
            if (
                "digital_encoding" in a.attrs
                and "dictionary" in a.attrs["digital_encoding"]
            ):
                de = a.attrs["digital_encoding"]
                data.append(
                    pa.DictionaryArray.from_arrays(
                        a.values,
                        de["dictionary"],
                        ordered=de.get("ordered", False),
                    )
                )
            else:
                data.append(pa.array(a.values))
        content = dict(zip(columns, data))
        content[self.dim_name] = self.index
        return pa.Table.from_pydict(content)

    def to_parquet(self, filename):
        import pyarrow.parquet as pq

        t = self.to_pyarrow()
        pq.write_table(t, filename)

    def to_pandas(self) -> pd.DataFrame:
        """
        Convert this dataset into a pandas DataFrame.

        The resulting DataFrame is always a copy of the data in the dataset.

        Returns
        -------
        pandas.DataFrame
        """
        return pd.DataFrame(self._to_pydict(), index=self.index, copy=True)

    def eval(
        self,
        expr: str,
        parser: str = "pandas",
        engine: str | None = None,
        local_dict=None,
        global_dict=None,
    ):
        """
        Evaluate a Python expression as a string using various backends.

        Parameters
        ----------
        expr : str
            The expression to evaluate. This string cannot contain any Python
            `statements
            <https://docs.python.org/3/reference/simple_stmts.html#simple-statements>`__,
            only Python `expressions
            <https://docs.python.org/3/reference/simple_stmts.html#expression-statements>`__.
        parser : {'pandas', 'python'}, default 'pandas'
            The parser to use to construct the syntax tree from the expression. The
            default of ``'pandas'`` parses code slightly different than standard
            Python. Alternatively, you can parse an expression using the
            ``'python'`` parser to retain strict Python semantics.  See the
            :ref:`enhancing performance <enhancingperf.eval>` documentation for
            more details.
        engine : {'python', 'numexpr'}, default 'numexpr'
            The engine used to evaluate the expression. Supported engines are
            - None : tries to use ``numexpr``, falls back to ``python``
            - ``'numexpr'`` : This default engine evaluates pandas objects using
              numexpr for large speed ups in complex expressions with large frames.
            - ``'python'`` : Performs operations as if you had ``eval``'d in top
              level python. This engine is generally not that useful.
        local_dict : dict or None, optional
            A dictionary of local variables, taken from locals() by default.
        global_dict : dict or None, optional
            A dictionary of global variables, taken from globals() by default.

        Returns
        -------
        DataArray or numeric scalar
        """
        result = pd.eval(
            expr,
            parser=parser,
            engine=engine,
            local_dict=local_dict,
            global_dict=global_dict,
            resolvers=[self.dataset],
        )
        if result.size == self.size:
            return DataArray(np.asarray(result), coords=self.dataset.coords)
        else:
            return result


@xr.register_dataarray_accessor("single_dim")
class _SingleDimArray:
    """Convenience accessor for single-dimension datasets."""

    __slots__ = ("dataarray", "dim_name")

    def __init__(self, dataarray: DataArray):
        self.dataarray = dataarray
        if len(self.dataarray.dims) != 1:
            raise ValueError("single_dim implies a single dimension dataset")
        self.dim_name = self.dataarray.dims[0]

    @property
    def coords(self):
        return self.dataarray.coords[self.dim_name]

    @property
    def index(self):
        return self.dataarray.indexes[self.dim_name]

    def rename(self, name: str) -> DataArray:
        """Rename the single dimension."""
        if self.dim_name == name:
            return self.dataarray
        return self.dataarray.rename({self.dim_name: name})

    def to_pandas(self) -> pd.Series:
        """
        Convert this array into a pandas Series.

        If this array is categorical (i.e. with a simple dictionary-based
        digital encoding) then the result will be a Series with categorical dtype.

        The DataArray's `name` attribute is preserved in the result.
        """
        if self.dataarray.cat.is_categorical():
            return pd.Series(
                pd.Categorical.from_codes(
                    self.dataarray,
                    self.dataarray.cat.categories,
                    self.dataarray.cat.ordered,
                ),
                index=self.index,
                name=self.dataarray.name,
            )
        else:
            result = self.dataarray.to_pandas()
            if self.dataarray.name:
                result = result.rename(self.dataarray.name)
            return result

    def to_pyarrow(self):
        if self.dataarray.cat.is_categorical():
            return pa.DictionaryArray.from_arrays(
                self.dataarray.data, self.dataarray.cat.categories
            )
        else:
            return pa.array(self.dataarray.data)


@xr.register_dataset_accessor("iloc")
class _iLocIndexer:
    """
    Purely integer-location based indexing for selection by position on 1-d Datasets.

    In many ways, a dataset with a single dimensions is like a pandas DataFrame,
    with the one dimension giving the rows, and the variables as columns. This
    analogy eventually breaks down (DataFrame columns are ordered, Dataset
    variables are not) but the similarities are enough that it’s sometimes
    convenient to have iloc functionality enabled. This only works for indexing
    on the rows, but if there’s only the one dimension the complexity of isel
    is not needed.
    """

    __slots__ = ("dataset",)

    def __init__(self, dataset: Dataset):
        self.dataset = dataset

    def __getitem__(self, key: Mapping[Hashable, Any]) -> Dataset:
        if not is_dict_like(key):
            if len(self.dataset.dims) == 1:
                dim_name = self.dataset.dims.__iter__().__next__()
                key = {dim_name: key}
            else:
                raise TypeError(
                    "can only lookup dictionaries from Dataset.iloc, "
                    "unless there is only one dimension"
                )
        return self.dataset.isel(key)


@xr.register_dataarray_accessor("iloc")
class _iLocArrayIndexer:
    """
    Purely integer-location based indexing for selection by position on 1-d DataArrays.

    In many ways, a dataset with a single dimensions is like a pandas DataFrame,
    with the one dimension giving the rows, and the variables as columns. This
    analogy eventually breaks down (DataFrame columns are ordered, Dataset
    variables are not) but the similarities are enough that it’s sometimes
    convenient to have iloc functionality enabled. This only works for indexing
    on the rows, but if there’s only the one dimension the complexity of isel
    is not needed.
    """

    __slots__ = ("dataarray",)

    def __init__(self, dataarray: DataArray):
        self.dataarray = dataarray

    def __getitem__(self, key: Mapping[Hashable, Any]) -> DataArray:
        if not is_dict_like(key):
            if len(self.dataarray.dims) == 1:
                dim_name = self.dataarray.dims.__iter__().__next__()
                key = {dim_name: key}
            else:
                raise TypeError(
                    "can only lookup dictionaries from DataArray.iloc, "
                    "unless there is only one dimension"
                )
        return self.dataarray.isel(key)


xr.Dataset.rename_dims_and_coords = xr.Dataset.rename


@register_dataset_method
def rename_or_ignore(self, dims_dict=None, **dims_kwargs):
    from xarray.core.utils import either_dict_or_kwargs

    dims_dict = either_dict_or_kwargs(dims_dict, dims_kwargs, "rename_dims_and_coords")
    dims_dict = {
        k: v for (k, v) in dims_dict.items() if (k in self.dims or k in self._variables)
    }
    return self.rename(dims_dict)


@register_dataset_method
def to_zarr_zip(self, *args, **kwargs):
    """
    Write dataset contents to a zarr group.

    Parameters
    ----------
    store : MutableMapping, str or Path, optional
        Store or path to directory in file system.  If given with a
        ".zarr.zip" extension, and keyword arguments limited to 'mode' and
        'compression', then a ZipStore will be created, populated, and then
        immediately closed.
    chunk_store : MutableMapping, str or Path, optional
        Store or path to directory in file system only for Zarr array chunks.
        Requires zarr-python v2.4.0 or later.
    mode : {"w", "w-", "a", None}, optional
        Persistence mode: "w" means create (overwrite if exists);
        "w-" means create (fail if exists);
        "a" means override existing variables (create if does not exist).
        If ``append_dim`` is set, ``mode`` can be omitted as it is
        internally set to ``"a"``. Otherwise, ``mode`` will default to
        `w-` if not set.
    synchronizer : object, optional
        Zarr array synchronizer.
    group : str, optional
        Group path. (a.k.a. `path` in zarr terminology.)
    encoding : dict, optional
        Nested dictionary with variable names as keys and dictionaries of
        variable specific encodings as values, e.g.,
        ``{"my_variable": {"dtype": "int16", "scale_factor": 0.1,}, ...}``
    compute : bool, optional
        If True write array data immediately, otherwise return a
        ``dask.delayed.Delayed`` object that can be computed to write
        array data later. Metadata is always updated eagerly.
    consolidated : bool, optional
        If True, apply zarr's `consolidate_metadata` function to the store
        after writing metadata.
    append_dim : hashable, optional
        If set, the dimension along which the data will be appended. All
        other dimensions on overriden variables must remain the same size.
    region : dict, optional
        Optional mapping from dimension names to integer slices along
        dataset dimensions to indicate the region of existing zarr array(s)
        in which to write this dataset's data. For example,
        ``{'x': slice(0, 1000), 'y': slice(10000, 11000)}`` would indicate
        that values should be written to the region ``0:1000`` along ``x``
        and ``10000:11000`` along ``y``.

        Two restrictions apply to the use of ``region``:

        - If ``region`` is set, _all_ variables in a dataset must have at
          least one dimension in common with the region. Other variables
          should be written in a separate call to ``to_zarr()``.
        - Dimensions cannot be included in both ``region`` and
          ``append_dim`` at the same time. To create empty arrays to fill
          in with ``region``, use a separate call to ``to_zarr()`` with
          ``compute=False``. See "Appending to existing Zarr stores" in
          the reference documentation for full details.
    compression : int, optional
        Only used for ".zarr.zip" files.  By default zarr uses blosc
        compression for chunks, so adding another layer of compression here
        is typically redundant.

    References
    ----------
    https://zarr.readthedocs.io/

    Notes
    -----
    Zarr chunking behavior:
        If chunks are found in the encoding argument or attribute
        corresponding to any DataArray, those chunks are used.
        If a DataArray is a dask array, it is written with those chunks.
        If not other chunks are found, Zarr uses its own heuristics to
        choose automatic chunk sizes.
    """
    if len(args) == 1 and isinstance(args[0], str) and args[0].endswith(".zarr.zip"):
        if {"compression", "mode"}.issuperset(kwargs.keys()):
            import zarr

            with zarr.ZipStore(args[0], **kwargs) as store:
                self.to_zarr(store)
            return
    return super().to_zarr(*args, **kwargs)


def _to_ast_literal(x):
    if isinstance(x, dict):
        return (
            "{"
            + ", ".join(
                f"{_to_ast_literal(k)}: {_to_ast_literal(v)}" for k, v in x.items()
            )
            + "}"
        )
    elif isinstance(x, list):
        return "[" + ", ".join(_to_ast_literal(i) for i in x) + "]"
    elif isinstance(x, tuple):
        return "(" + ", ".join(_to_ast_literal(i) for i in x) + ")"
    elif isinstance(x, pd.Index):
        return _to_ast_literal(x.to_list())
    elif isinstance(x, np.ndarray):
        return _to_ast_literal(list(x))
    elif isinstance(x, np.str_):
        return repr(str(x))
    else:
        return repr(x)


def _to_evalable_string(x):
    if x is None:
        return " < None > "
    elif x is True:
        return " < True > "
    elif x is False:
        return " < False > "
    else:
        return f" {_to_ast_literal(x)} "


def _from_evalable_string(x):
    if isinstance(x, str):
        # if x.startswith(" {") and x.endswith("} "):
        #     return ast.literal_eval(x[1:-1])
        if x == " < None > ":
            return None
        if x == " < True > ":
            return True
        if x == " < False > ":
            return False
        if x.startswith(" ") and x.endswith(" "):
            try:
                return ast.literal_eval(x.strip(" "))
            except Exception:
                print(x)
                raise
    else:
        return x


@register_dataset_method
def to_zarr_with_attr(self, *args, **kwargs):
    """
    Write dataset contents to a zarr group.

    Parameters
    ----------
    store : MutableMapping, str or Path, optional
        Store or path to directory in file system.  If given with a
        ".zarr.zip" extension, and keyword arguments limited to 'mode' and
        'compression', then a ZipStore will be created, populated, and then
        immediately closed.
    chunk_store : MutableMapping, str or Path, optional
        Store or path to directory in file system only for Zarr array chunks.
        Requires zarr-python v2.4.0 or later.
    mode : {"w", "w-", "a", None}, optional
        Persistence mode: "w" means create (overwrite if exists);
        "w-" means create (fail if exists);
        "a" means override existing variables (create if does not exist).
        If ``append_dim`` is set, ``mode`` can be omitted as it is
        internally set to ``"a"``. Otherwise, ``mode`` will default to
        `w-` if not set.
    synchronizer : object, optional
        Zarr array synchronizer.
    group : str, optional
        Group path. (a.k.a. `path` in zarr terminology.)
    encoding : dict, optional
        Nested dictionary with variable names as keys and dictionaries of
        variable specific encodings as values, e.g.,
        ``{"my_variable": {"dtype": "int16", "scale_factor": 0.1,}, ...}``
    compute : bool, optional
        If True write array data immediately, otherwise return a
        ``dask.delayed.Delayed`` object that can be computed to write
        array data later. Metadata is always updated eagerly.
    consolidated : bool, optional
        If True, apply zarr's `consolidate_metadata` function to the store
        after writing metadata.
    append_dim : hashable, optional
        If set, the dimension along which the data will be appended. All
        other dimensions on overriden variables must remain the same size.
    region : dict, optional
        Optional mapping from dimension names to integer slices along
        dataset dimensions to indicate the region of existing zarr array(s)
        in which to write this dataset's data. For example,
        ``{'x': slice(0, 1000), 'y': slice(10000, 11000)}`` would indicate
        that values should be written to the region ``0:1000`` along ``x``
        and ``10000:11000`` along ``y``.

        Two restrictions apply to the use of ``region``:

        - If ``region`` is set, _all_ variables in a dataset must have at
          least one dimension in common with the region. Other variables
          should be written in a separate call to ``to_zarr()``.
        - Dimensions cannot be included in both ``region`` and
          ``append_dim`` at the same time. To create empty arrays to fill
          in with ``region``, use a separate call to ``to_zarr()`` with
          ``compute=False``. See "Appending to existing Zarr stores" in
          the reference documentation for full details.
    compression : int, optional
        Only used for ".zarr.zip" files.  By default zarr uses blosc
        compression for chunks, so adding another layer of compression here
        is typically redundant.

    References
    ----------
    https://zarr.readthedocs.io/

    Notes
    -----
    Zarr chunking behavior:
        If chunks are found in the encoding argument or attribute
        corresponding to any DataArray, those chunks are used.
        If a DataArray is a dask array, it is written with those chunks.
        If not other chunks are found, Zarr uses its own heuristics to
        choose automatic chunk sizes.
    """
    obj = self.copy()
    for k in self:
        attrs = {}
        for aname, avalue in self[k].attrs.items():
            attrs[aname] = _to_evalable_string(avalue)
        obj[k] = self[k].assign_attrs(attrs)
    if hasattr(self, "coords"):
        for k in self.coords:
            attrs = {}
            for aname, avalue in self.coords[k].attrs.items():
                attrs[aname] = _to_evalable_string(avalue)
            obj.coords[k] = self.coords[k].assign_attrs(attrs)
    attrs = {}
    for aname, avalue in self.attrs.items():
        attrs[aname] = _to_evalable_string(avalue)
    obj = obj.assign_attrs(attrs)
    return obj.to_zarr(*args, **kwargs)


@register_dataset_method
def to_table(self):
    """
    Convert dataset contents to a pyarrow Table.

    This dataset must not contain more than one dimension.
    """
    assert isinstance(self, Dataset)
    if len(self.dims) != 1:
        raise ValueError("Only 1-dim datasets can be converted to tables")

    import pyarrow as pa

    from .relationships import sparse_array_type

    def to_numpy(var):
        """Coerces wrapped data to numpy and returns a numpy.ndarray."""
        data = var.data
        if hasattr(data, "chunks"):
            data = data.compute()
        if isinstance(data, sparse_array_type):
            data = data.todense()
        return np.asarray(data)

    pydict = {}
    for i in self.variables:
        dictionary = self[i].attrs.get("DICTIONARY", None)
        if dictionary is not None:
            pydict[i] = pa.DictionaryArray.from_arrays(
                to_numpy(self[i]),
                dictionary,
            )
        else:
            pydict[i] = pa.array(to_numpy(self[i]))
    return pa.Table.from_pydict(pydict)


@register_dataset_method
def select_and_rename(self, name_dict=None, **names):
    """
    Select and rename variables from this Dataset.

    Parameters
    ----------
    name_dict, **names: dict
        The keys or keyword arguments give the current names of the
        variables that will be selected out of this Dataset.  The values
        give the new names of the same variables in the resulting Dataset.

    Returns
    -------
    Dataset
    """
    if name_dict is None:
        name_dict = names
    else:
        name_dict.update(names)
    return self[list(name_dict.keys())].rename(name_dict)


@register_dataset_method
def max_float_precision(self, p=32):
    """
    Set the maximum precision for floating point values.

    This modifies the Dataset in-place.

    Parameters
    ----------
    p : {64, 32, 16}
        The max precision to set.

    Returns
    -------
    self
    """
    for i in self:
        if np.issubdtype(self[i].dtype, np.floating):
            if self[i].dtype.itemsize > p / 8:
                self[i] = self[i].astype(f"float{p}")
    return self


@register_dataset_method
def interchange_dims(self, dim1, dim2):
    """
    Rename a pair of dimensions by swapping their names.

    Parameters
    ----------
    dim1, dim2 : str
        The names of the two dimensions to swap.

    Returns
    -------
    Dataset
    """
    p21 = "PLACEHOLD21"
    p12 = "PLACEHOLD12"
    s1 = {dim1: p12, dim2: p21}
    s2 = {p12: dim2, p21: dim1}
    rv = {}
    vr = {}
    if dim1 in self.variables:
        rv[dim1] = p12
        vr[p12] = dim2
    if dim2 in self.variables:
        rv[dim2] = p21
        vr[p21] = dim1
    return self.rename_dims(s1).rename_vars(rv).rename_dims(s2).rename_vars(vr)


def from_named_objects(*args):
    """
    Create a Dataset by populating it with named objects.

    A mapping of names to values is first created, and then that mapping is
    used in the standard constructor to initialize a Dataset.

    Parameters
    ----------
    *args : Any
        A collection of objects, each exposing a `name` attribute.

    Returns
    -------
    Dataset
    """
    objs = {}
    for n, a in enumerate(args):
        try:
            name = a.name
        except AttributeError:
            raise ValueError(f"argument {n} has no name") from None
        if name is None:
            raise ValueError(f"the name for argument {n} is None")
        objs[name] = np.asarray(a)
    return xr.Dataset(objs)


@register_dataset_method
def ensure_integer(dataset, names, bitwidth=32, inplace=False):
    """
    Convert dataset variables to integers, if they are not already integers.

    Parameters
    ----------
    names : Iterable[str]
        Variable names in this dataset to convert.
    bitwidth : int, default 32
        Bit width of integers that are created when a conversion is made.
        Note that variables that are already integer are not modified,
        even if their bit width differs from this.
    inplace : bool, default False
        Whether to make the conversion in-place on this Dataset, or
        return a copy.

    Returns
    -------
    Dataset
    """
    if inplace:
        result = dataset
    else:
        result = dataset.copy()
    for name in names:
        if name not in result:
            continue
        if not np.issubdtype(result[name].dtype, np.integer):
            result[name] = result[name].astype(f"int{bitwidth}")
    if not inplace:
        return result


def filter_name_tokens(expr, matchable_names=None):
    name_tokens = extract_all_name_tokens(expr)
    name_tokens -= {"_args", "_inputs", "_outputs", "np"}
    name_tokens -= well_known_names
    if matchable_names:
        name_tokens &= matchable_names
    return name_tokens


def _dyno(k, v):
    if isinstance(v, str) and v[0] == "@":
        return f"__dynamic_{k}{v}"
    elif v is None:
        return f"__dynamic_{k}"
    else:
        return v


def _flip_flop_def(v):
    if "# sharrow:" in v:
        return v.split("# sharrow:", 1)[1].strip()
    else:
        return v
